import freeorion as fo
import random

import universe_statistics
from names import get_name_list, random_name
from options import (
    HS_ACCEPTABLE_PLANET_TYPES,
    HS_MAX_JUMP_DISTANCE_LIMIT,
    HS_MIN_DISTANCE_PRIORITY_LIMIT,
    HS_MIN_PLANET_TYPES_IN_VICINITY,
    HS_MIN_PLANETS_IN_VICINITY_PER_SYSTEM,
    HS_MIN_PLANETS_IN_VICINITY_TOTAL,
    HS_MIN_SYSTEMS_IN_VICINITY,
    HS_VICINITY_RANGE,
)
from planets import (
    calc_planet_size,
    calc_planet_type,
    planet_sizes_real,
    planet_types,
    planet_types_real,
)
from starsystems import pick_star_type, star_types_real
from util import report_error, unique_product


def get_empire_name_generator():
    """
    String generator, return random empire name from string list,
    if string list is empty generate random name.
    """
    empire_names = get_name_list("EMPIRE_NAMES")
    random.shuffle(empire_names)
    while True:
        if empire_names:
            yield empire_names.pop()
        else:
            yield random_name(5)


# generate names for empires, use next(empire_name_generator) to get next name.
empire_name_generator = get_empire_name_generator()


def get_starting_species_pool():
    """
    Empire species pool generator, return random empire species and ensure somewhat even distribution
    """
    # fill the initial pool of playable species, without repetitions unless RULE_ALLOW_REPEATED_SPECIES is true
    if not fo.getGameRules().getToggle("RULE_ALLOW_REPEATED_SPECIES"):
        pool = fo.get_playable_species()
    else:
        pool = fo.get_playable_species() * 2

    # randomize order in initial pool so we don't get the same species all the time
    random.shuffle(pool)
    # generator loop
    while True:
        # if our pool is exhausted (because we have more players than species instances in our initial pool)
        # refill the pool with one set of all playable species
        if not pool:
            pool = fo.get_playable_species()
            # again, randomize order in refilled pool so we don't get the same species all the time
            random.shuffle(pool)
        # pick and return next species, and remove it from our pool
        yield pool.pop()


# generates starting species for empires, use next(starting_species_pool) to get next species
starting_species_pool = get_starting_species_pool()


def count_planets_in_systems(systems, planet_types_filter=HS_ACCEPTABLE_PLANET_TYPES):
    """
    Return the total number of planets in the specified group of systems,
    only count the planet types specified in planet_types_filter.
    """
    num_planets = 0
    for system in systems:
        num_planets += len([p for p in fo.sys_get_planets(system) if fo.planet_get_type(p) in planet_types_filter])
    return num_planets


def list_planet_types_in_systems(systems):
    """
    Get a list of planet types in systems
    """
    planet_types_in_systems = []
    for system in systems:
        for planet in fo.sys_get_planets(system):
            planet_types_in_systems.append(fo.planet_get_type(planet))
    return planet_types_in_systems


def count_planet_types_in_systems(systems):
    """
    Return the total number of planet types in the specified group of systems.
    """
    planet_types_in_systems = list_planet_types_in_systems(systems)
    num_planet_types = len(set(planet_types_in_systems))
    return num_planet_types


def calculate_home_system_merit(system):
    """Calculate the system's merit as the number of planets within HS_VICINTIY_RANGE."""
    return count_planets_in_systems(fo.systems_within_jumps_unordered(HS_VICINITY_RANGE, [system]))


def min_planets_in_vicinity_limit(num_systems):
    """
    Calculates the minimum planet limit for the specified number of systems.
    This limit is the lower of HS_MIN_PLANETS_IN_VICINITY_TOTAL or HS_MIN_PLANETS_IN_VICINITY_PER_SYSTEM
    planets per system.
    """
    return min(HS_MIN_PLANETS_IN_VICINITY_TOTAL, num_systems * HS_MIN_PLANETS_IN_VICINITY_PER_SYSTEM)


class HomeSystemFinder:
    """Finds a set of home systems with a least ''num_home_systems'' systems."""

    def __init__(self, _num_home_systems):
        # cache of sytem merits
        self.system_merit = {}
        self.num_home_systems = _num_home_systems

    def find_home_systems_for_min_jump_distance(self, systems_pool, min_jumps):  # noqa: C901
        """
        Return a good list of home systems or an empty list if there are fewer than num_home_systems in the pool.

        A good list of home systems are at least the specified minimum number of jumps apart,
        with the best minimum system merit of all such lists picked randomly from the ''systems_pool''.

        Algorithm:
        Make several attempts to find systems that match the condition
        of being at least min_jumps apart.
        Use the minimum merit of the best num_home_system systems found
        to compare the candidate with the current best set of systems.
        On each attempt use the minimum merit of the current best set of home
        systems to truncate the pool of candidates.
        """

        # precalculate the system merits
        for system in systems_pool:
            if system not in self.system_merit:
                self.system_merit[system] = calculate_home_system_merit(system)

        # The list of merits and systems sorted in descending order by merit.
        all_merit_system = sorted([(self.system_merit[s], s) for s in systems_pool], reverse=True)

        current_merit_lower_bound = 0
        best_candidate = []

        # Cap the number of attempts when the found number of systems is less than the target
        # num_home_systems because this indicates that the min_jumps is too large and/or the
        # systems_pool is too small to ever succeed.

        # From experimentation with cluster and 3 arm spiral galaxies, with low, med and high
        # starlane density and (number of systems, number of home systems) pairs of (9999, 399),
        # (999, 39) and (199, 19) the following was observered.  The distribution of candidate
        # length is a normal random variable with standard deviation approximately equal to

        # expected_len_candidate_std = (len(systems) ** (1.0/2.0)) * 0.03

        # which is about 1 for 1000 systems.  It is likely that anylen(candidate) is within 1
        # standard deviation of the expected len(candidate)

        # If we are within the MISS_THRESHOLD of the target then try up to num_complete_misses more times.
        MISS_THRESHOLD = 3
        num_complete_misses_remaining = 4

        # Cap the number of attempts to the smaller of the number of systems in the pool, or 100
        attempts = min(100, len(systems_pool))

        while attempts and num_complete_misses_remaining:
            # use a local pool of all candidate systems better than the worst threshold merit
            all_merit_system = [(m, s) for (m, s) in all_merit_system if m > current_merit_lower_bound]
            local_pool = {s for (m, s) in all_merit_system}

            if len(local_pool) < self.num_home_systems:
                if not best_candidate:
                    print(
                        "Failing in find_home_systems_for_min_jump_distance because "
                        "current_merit_lower_bound = {} trims local pool to {} systems "
                        "which is less than num_home_systems {}.".format(
                            current_merit_lower_bound, len(local_pool), self.num_home_systems
                        )
                    )
                break

            attempts = min(attempts - 1, len(local_pool))

            candidate = []
            while local_pool:
                member = random.choice(list(local_pool))
                candidate.append(member)

                # remove all neighbors from the local pool
                local_pool -= set(fo.systems_within_jumps_unordered(min_jumps, [member]))

            # Count complete misses when number of candidates is not close to the target.
            if len(candidate) < (self.num_home_systems - MISS_THRESHOLD):
                num_complete_misses_remaining -= 1

            if len(candidate) < self.num_home_systems:
                continue

            # Calculate the merit of the current attempt.  If it is the best so far
            # keep it and update the merit_threshold
            merit_system = sorted([(self.system_merit[s], s) for s in candidate], reverse=True)[: self.num_home_systems]

            (merit, system) = merit_system[-1]

            # If we have a better candidate, set the new lower bound and try for a better candidate.
            if merit > current_merit_lower_bound:
                print(f"Home system set merit lower bound improved from {current_merit_lower_bound} to " f"{merit}")
                current_merit_lower_bound = merit
                best_candidate = [s for (_, s) in merit_system]

                # Quit successfully if the lowest merit system meets the minimum threshold
                if merit >= min_planets_in_vicinity_limit(
                    len(fo.systems_within_jumps_unordered(HS_VICINITY_RANGE, [system]))
                ):
                    break

        return best_candidate


def find_home_systems(num_home_systems, pool_list, jump_distance, min_jump_distance):
    """
    Tries to find a specified number of home systems which are as far apart from each other as possible.
    Starts with the specified jump distance and reduces that limit until enough systems can be found or the
    jump distance drops below the specified minimum jump distance limit (in this case fail).
    For each jump distance several attempts are made: one for each pool passed in pool_list.
    This parameter contains a list of tuples, each tuple has a pool of systems as first element and a description
    of the pool for logging purposes as second element.
    """

    finder = HomeSystemFinder(num_home_systems)
    # try to find home systems, decrease the min jumps until enough systems can be found, or the jump distance drops
    # below the specified minimum jump distance (which means failure)
    while jump_distance >= min_jump_distance:
        print("Trying to find", num_home_systems, "home systems that are at least", jump_distance, "jumps apart...")

        # try to pick our home systems by iterating over the pools we got
        for pool, pool_label in pool_list:
            print("...use", pool_label)

            # check if the pool has enough systems to pick from
            if len(pool) <= num_home_systems:
                # no, the pool has less systems than home systems requested, so just skip trying using that pool
                print("...pool only has", len(pool), "systems, skip trying to use it")
                continue

            # try to pick home systems
            home_systems = finder.find_home_systems_for_min_jump_distance(pool, jump_distance)
            # check if we got enough
            if len(home_systems) >= num_home_systems:
                # yes, we got what we need, return the home systems we found
                print("...", len(home_systems), "systems found")
                return home_systems
            else:
                # no, try next pool
                print("...only", len(home_systems), "systems found")

        # we did not find enough home systems with the current jump distance requirement,
        # so decrease the jump distance and try again
        jump_distance -= 1

    # all attempts came up with too few systems, return empty list to indicate failure
    return []


def add_planets_to_vicinity(vicinity, num_planets, acceptable_planet_types, allow_repeat_types, gsd):  # noqa C901
    """
    Adds the specified number of planets to the specified systems.
    """
    print("Adding", num_planets, "planets to the following systems:", vicinity)

    # first, compile a list containing all the free orbits in the specified systems
    # begin with adding the free orbits of all systems that have a real star (that is, no neutron star, black hole,
    # and not no star), if that isn't enough, also one, by one, add the free orbits of neutron star, black hole and
    # no star systems (in that order) until we have enough free orbits

    # for that, we use this list of tuples
    # the first tuple contains all real star types, the following tuples the neutron, black hole and no star types,
    # so we can iterate over this list and only add the free orbits of systems that match the respective star types
    # each step
    # this way we can prioritize the systems we want to add planets to by star type
    acceptable_star_types_list = [
        star_types_real,
        (fo.starType.noStar,),
        (fo.starType.neutron,),
        (fo.starType.blackHole,),
    ]

    # store the free orbits as a list of tuples of (system, orbit)
    free_orbits_map = []

    # now, iterate over the list of acceptable star types
    for acceptable_star_types in acceptable_star_types_list:
        # check all systems in the list of systems we got passed into this function
        for system in vicinity:
            # if this system has a star type we want to accept in this step, add its free orbits to our list
            if fo.sys_get_star_type(system) in acceptable_star_types:
                free_orbits_map.extend([(system, orbit) for orbit in fo.sys_free_orbits(system)])
        # check if we got enough free orbits after completing this step
        # we want 4 times as much free orbits as planets we want to add, that means each system shouldn't get more
        # than 2-3 additional planets on average
        if len(free_orbits_map) > (num_planets * 4):
            break

    # if we got less free orbits than planets that should be added, something is wrong
    # in that case abort and log an error
    if len(free_orbits_map) < num_planets:
        report_error("Python add_planets_to_vicinity: less free orbits than planets to add - cancelled")

    print("...free orbits available:", free_orbits_map)
    # as we will pop the free orbits from this list afterwards, shuffle it to randomize the order of the orbits
    random.shuffle(free_orbits_map)

    # add the requested number of planets
    while num_planets > 0:
        # get the next free orbit from the list we just compiled
        system, orbit = free_orbits_map.pop()

        # check the star type of the system containing the orbit we got
        star_type = fo.sys_get_star_type(system)
        if star_type in [fo.starType.noStar, fo.starType.blackHole]:
            # if it is a black hole or has no star, change the star type
            # pick a star type, continue until we get a real star
            # don't accept neutron, black hole or no star
            print("...system picked to add a planet has star type", star_type)
            while star_type not in star_types_real:
                star_type = pick_star_type(gsd.age)
            print("...change that to", star_type)
            fo.sys_set_star_type(system, star_type)

        # pick a planet size and type, continue until we get a type from acceptable_planet_types
        planet_type = fo.planetType.unknown
        while planet_type not in acceptable_planet_types:
            planet_size = calc_planet_size(star_type, orbit, fo.galaxySetupOptionGeneric.high, gsd.shape)
            planet_type = calc_planet_type(star_type, orbit, planet_size)

        # finally, create the planet in the system and orbit we got
        print("...adding", planet_size, planet_type, "planet to system", system)
        if fo.create_planet(planet_size, planet_type, system, orbit, "") == fo.invalid_object():
            report_error("Python add_planets_to_vicinity: create planet in system %d failed" % system)

        if not allow_repeat_types:
            acceptable_planet_types.remove(planet_type)

        # continue with next planet
        num_planets -= 1
        # safety check for empty set
        if not acceptable_planet_types:
            num_planets = 0
            print("planet types set is empty")


def compile_home_system_list(num_home_systems, systems, gsd):  # noqa: C901
    """
    Compiles a list with a requested number of home systems.
    """
    print("Compile home system list:", num_home_systems, "systems requested")

    # if the list of systems to choose home systems from is empty, report an error and return empty list
    if not systems:
        report_error("Python generate_home_system_list: no systems to choose from")
        return []

    # calculate an initial minimal number of jumps that the home systems should be apart,
    # based on the total number of systems to choose from and the requested number of home systems
    # don't let min_jumps be either:
    # a.) larger than a defined limit, because an unreasonably large number is really not at all needed,
    #     and with large galaxies an excessive amount of time can be used in failed attempts
    # b.) lower than the minimum jump distance limit that should be considered high priority (see options.py),
    #     otherwise no attempt at all would be made to enforce the other requirements for home systems (see below)
    min_jumps = min(
        HS_MAX_JUMP_DISTANCE_LIMIT, max(int(len(systems) / (num_home_systems * 2)), HS_MIN_DISTANCE_PRIORITY_LIMIT)
    )

    # home systems must have a certain minimum of systems and planets in their near vicinity
    # we will try to select our home systems from systems that match this criteria, if that fails, we will select our
    # home systems from all systems and add the missing number planets to the systems in their vicinity afterwards
    # the minimum system and planet limit and the jump range that defines the "near vicinity" are controlled by the
    # HS_* option constants in options.py (see there)

    # we start by building three additional pools of systems: one that contains all systems that match the criteria
    # completely (minimum systems, minimum planets and minimum planet types), one that matches 2 of the criteria
    # (meets the min systems and planets limit), and one that contains all systems that match the criteria at least
    # partially (meets the min systems limit).
    pool_matching_sys_and_planet_and_planet_type_limit = []
    pool_matching_sys_and_planet_limit = []
    pool_matching_sys_limit = []
    for system in systems:
        systems_in_vicinity = fo.systems_within_jumps_unordered(HS_VICINITY_RANGE, [system])
        if len(systems_in_vicinity) >= HS_MIN_SYSTEMS_IN_VICINITY:
            pool_matching_sys_limit.append(system)
            if count_planets_in_systems(systems_in_vicinity) >= min_planets_in_vicinity_limit(len(systems_in_vicinity)):
                pool_matching_sys_and_planet_limit.append(system)
                if (
                    count_planets_in_systems(systems_in_vicinity)
                    >= min_planets_in_vicinity_limit(len(systems_in_vicinity))
                    and count_planet_types_in_systems(systems_in_vicinity) >= HS_MIN_PLANET_TYPES_IN_VICINITY
                ):
                    pool_matching_sys_and_planet_and_planet_type_limit.append(system)
    print(
        len(pool_matching_sys_and_planet_and_planet_type_limit),
        "systems meet the min systems, planets and planet types in the near vicinity limit",
    )
    print(
        len(pool_matching_sys_and_planet_limit), "systems meet the min systems and planets in the near vicinity limit"
    )
    print(len(pool_matching_sys_limit), "systems meet the min systems in the near vicinity limit")

    # now try to pick the requested number of home systems
    # we will do this by calling find_home_systems, which takes a list of tuples defining the pools from which to pick
    # the home systems; it will use the pools in the order in which they appear in the list, so put better pools first

    # we will make two attempts: the first one with the filtered pools we just created, and tell find_home_systems
    # to start with the min_jumps jumps distance we calculated above, but not to go lower than
    # HS_MIN_DISTANCE_PRIORITY_LIMIT

    print("First attempt: trying to pick home systems from the filtered pools of preferred systems")
    pool_list = [
        # the best pool is of course the one where all systems meet all 3 criteria
        (
            pool_matching_sys_and_planet_and_planet_type_limit,
            "pool of systems that meet min systems, planets and planet type limit",
        ),
        # the next best pool is of course the one where all systems meet BOTH the min systems and planets limit
        (pool_matching_sys_and_planet_limit, "pool of systems that meet both the min systems and planets limit"),
        # next the less preferred pool where all systems at least meets the min systems limit
        # specify 0 as number of requested home systems to pick as much systems as possible
        (pool_matching_sys_limit, "pool of systems that meet at least the min systems limit"),
    ]
    home_systems = find_home_systems(num_home_systems, pool_list, min_jumps, HS_MIN_DISTANCE_PRIORITY_LIMIT)

    # check if the first attempt delivered a list with enough home systems
    # if not, we make our second attempt, this time disregarding the filtered pools and using all systems, starting
    # again with the min_jumps jump distance limit and specifying 0 as number of required home systems to pick as much
    # systems as possible
    if len(home_systems) < num_home_systems:
        print("Second attempt: trying to pick home systems from all systems")
        home_systems = find_home_systems(num_home_systems, [(systems, "complete pool")], min_jumps, 1)

    # check if the selection process delivered a list with enough home systems
    # if not, our galaxy obviously is too crowded, report an error and return an empty list
    if len(home_systems) < num_home_systems:
        report_error(
            "Python generate_home_system_list: requested %d homeworlds in a galaxy with %d systems"
            % (num_home_systems, len(systems))
        )
        return []

    # check if we got more home systems than we requested
    if len(home_systems) > num_home_systems:
        # yes: calculate the number of planets in the near vicinity of each system
        # and store that value with each system in a map
        hs_planets_in_vicinity_map = {s: calculate_home_system_merit(s) for s in home_systems}
        # sort the home systems by the number of planets in their near vicinity using the map
        # now only pick the number of home systems we need, taking those with the highest number of planets
        home_systems = sorted(home_systems, key=hs_planets_in_vicinity_map.get, reverse=True)[:num_home_systems]

    # make sure all our home systems have a "real" star (that is, a star that is not a neutron star, black hole,
    # or even no star at all) and at least one planet in it
    for home_system in home_systems:
        # if this home system has no "real" star, change star type to a randomly selected "real" star
        if fo.sys_get_star_type(home_system) not in star_types_real:
            star_type = random.choice(star_types_real)
            print(
                "Home system",
                home_system,
                "has star type",
                fo.sys_get_star_type(home_system),
                ", changing that to",
                star_type,
            )
            fo.sys_set_star_type(home_system, star_type)

        # if this home system has no planets, create one in a random orbit
        # we take random values for type and size, as these will be set to suitable values later
        if not fo.sys_get_planets(home_system):
            print("Home system", home_system, "has no planets, adding one")
            planet = fo.create_planet(
                random.choice(planet_sizes_real),
                random.choice(planet_types_real),
                home_system,
                random.randint(0, fo.sys_get_num_orbits(home_system) - 1),
                "",
            )
            # if we couldn't create the planet, report an error and return an empty list
            if planet == fo.invalid_object():
                report_error("Python generate_home_system_list: couldn't create planet in home system")
                return []

    # finally, check again if all home systems meet the criteria of having the required minimum number of planets
    # within their near vicinity, if not, add the missing number of planets
    print("Checking if home systems have the required minimum of planets within the near vicinity...")
    for home_system in home_systems:
        # calculate the number of missing planets, and add them if this number is > 0
        systems_in_vicinity = fo.systems_within_jumps_unordered(HS_VICINITY_RANGE, [home_system])
        num_systems_in_vicinity = len(systems_in_vicinity)
        num_planets_in_vicinity = count_planets_in_systems(systems_in_vicinity)
        num_planets_to_add = min_planets_in_vicinity_limit(num_systems_in_vicinity) - num_planets_in_vicinity
        num_planet_types_in_vicinity = count_planet_types_in_systems(systems_in_vicinity)
        num_planet_types_to_add = HS_MIN_PLANET_TYPES_IN_VICINITY - num_planet_types_in_vicinity
        print(
            "Home system",
            home_system,
            "has",
            num_systems_in_vicinity,
            "systems,",
            num_planet_types_in_vicinity,
            "planet types, required minimum:",
            HS_MIN_PLANET_TYPES_IN_VICINITY,
            "and",
            num_planets_in_vicinity,
            "planets in the near vicinity, required minimum:",
            min_planets_in_vicinity_limit(num_systems_in_vicinity),
        )
        planet_types_in_systems = list_planet_types_in_systems(systems_in_vicinity)
        if num_planet_types_to_add > 0:
            planet_types_in_systems = list_planet_types_in_systems(systems_in_vicinity)
            # get the set of missing planet types
            planet_types_not_in_systems = set(planet_types).difference(planet_types_in_systems)
            systems_in_vicinity.remove(home_system)  # don't add planets to the home system, so remove it from the list
            add_planets_to_vicinity(
                sorted(systems_in_vicinity), num_planet_types_to_add, planet_types_not_in_systems, False, gsd
            )
            num_planets_to_add -= num_planet_types_to_add
            systems_in_vicinity.append(home_system)

        if num_planets_to_add > 0:
            systems_in_vicinity.remove(home_system)  # don't add planets to the home system, so remove it from the list
            # sort the systems_in_vicinity before adding, since the C++ engine doesn't guarrantee the same
            # platform independence as python.
            add_planets_to_vicinity(
                sorted(systems_in_vicinity), num_planets_to_add, HS_ACCEPTABLE_PLANET_TYPES, True, gsd
            )

    # as we've sorted the home system list before, lets shuffle it to ensure random order and return
    random.shuffle(home_systems)
    return home_systems


# noqa: C901
def setup_empire(empire, empire_name, home_system, starting_species, player_name, gsd):  # noqa: C901
    """
    Sets up various aspects of an empire, like empire name, homeworld, etc.
    """

    # set empire name, if no one is given, pick one randomly
    if not empire_name:
        print("No empire name set for player", player_name, ", picking one randomly")
        empire_name = next(empire_name_generator)
    fo.empire_set_name(empire, empire_name)
    print("Empire name for player", player_name, "is", empire_name)

    # check starting species, if no one is given, pick one randomly
    if starting_species == "RANDOM" or not starting_species:
        print("Picking random starting species for player", player_name)
        starting_species = next(starting_species_pool)
    print("Starting species for player", player_name, "is", starting_species)
    universe_statistics.empire_species[starting_species] += 1

    # If game rule "RULE_ENSURE_HABITABLE_PLANET_HW_VICINITY" is set check for an adequate or better planet and
    # if missing add one
    if fo.getGameRules().getToggle("RULE_ENSURE_HABITABLE_PLANET_HW_VICINITY"):
        systems_in_vicinity = fo.systems_within_jumps_unordered(HS_VICINITY_RANGE, [home_system])
        planet_types_in_vicinity = list_planet_types_in_systems(systems_in_vicinity)
        acceptable_planet_types = set()
        for planet_type in planet_types:
            # if this planet type is an adequate or good environment for the species, add it to the set of acceptable planet types
            if (
                fo.species_get_planet_environment(starting_species, planet_type) == fo.planetEnvironment.good
                or fo.species_get_planet_environment(starting_species, planet_type) == fo.planetEnvironment.adequate
            ):
                acceptable_planet_types.add(planet_type)
        if not set(planet_types_in_vicinity).intersection(acceptable_planet_types):
            systems_in_vicinity.remove(home_system)  # don't add planets to the home system, so remove it from the list
            print("Adding a habitable planet")
            add_planets_to_vicinity(sorted(systems_in_vicinity), 1, acceptable_planet_types, True, gsd)

    # pick a planet from the specified home system as homeworld
    planet_list = fo.sys_get_planets(home_system)
    # if the system is empty, report an error and return false, indicating failure
    if not planet_list:
        report_error("Python setup_empire: got home system with no planets")
        return False
    homeworld = random.choice(planet_list)

    # set selected planet as empire homeworld with selected starting species
    fo.empire_set_homeworld(empire, homeworld, starting_species)

    # set homeworld focus
    # check if the preferred focus for the starting species is among
    # the foci available on the homeworld planet
    available_foci = fo.planet_available_foci(homeworld)
    preferred_focus = fo.species_preferred_focus(starting_species)
    if preferred_focus in available_foci:
        # if yes, set the homeworld focus to the preferred focus
        print("Player", player_name, ": setting preferred focus", preferred_focus, "on homeworld")
        fo.planet_set_focus(homeworld, preferred_focus)
    elif len(available_foci) > 0:
        # if no, and there is at least one available focus,
        # just take the first of the list
        if preferred_focus == "":
            print(
                "Player",
                player_name,
                ": starting species",
                starting_species,
                "has no preferred focus, using",
                available_foci[0],
                "instead",
            )
        else:
            print(
                "Player",
                player_name,
                ": preferred focus",
                preferred_focus,
                "for starting species",
                starting_species,
                "not available on homeworld, using",
                available_foci[0],
                "instead",
            )
        fo.planet_set_focus(homeworld, available_foci[0])
    else:
        # if no focus is available on the homeworld, don't set any focus
        print("Player", player_name, ": no available foci on homeworld for starting species", starting_species)

    # give homeworld starting buildings
    print("Player", player_name, ": add starting buildings to homeworld")
    for item in fo.load_starting_buildings():
        fo.create_building(item.name, homeworld, empire)

    # unlock starting techs, buildings, hulls, ship parts, etc.
    # use default content file
    print("Player", player_name, ": add unlocked items")
    for item in fo.load_unlockable_item_list():
        fo.empire_unlock_item(empire, item.type, item.name)

    # add premade ship designs to empire
    print("Player", player_name, ": add premade ship designs")
    for ship_design in fo.design_get_premade_list():
        fo.empire_add_ship_design(empire, ship_design)

    # add starting fleets to empire
    # use default content file
    print("Player", player_name, ": add starting fleets")
    fleet_plans = fo.load_fleet_plan_list()
    for fleet_plan in fleet_plans:
        # should fleet be aggressive? check if any ships are armed.
        should_be_aggressive = False
        for ship_design in fleet_plan.ship_designs():
            design = fo.getPredefinedShipDesign(ship_design)
            if design is None:
                report_error("Looked up null design with name %s", ship_design)
            elif design.isArmed:
                should_be_aggressive = True
                break

        # first, create the fleet
        fleet = fo.create_fleet(fleet_plan.name(), home_system, empire, should_be_aggressive)
        # if the fleet couldn't be created, report an error and try to continue with the next fleet plan
        if fleet == fo.invalid_object():
            report_error("Python setup empire: couldn't create fleet %s" % fleet_plan.name())
            continue

        # second, iterate over the list of ship design names in the fleet plan
        for ship_design in fleet_plan.ship_designs():
            # create a ship in the fleet
            # if the ship couldn't be created, report an error and try to continue with the next ship design
            if fo.create_ship("", ship_design, starting_species, fleet) == fo.invalid_object():
                report_error(
                    f"Python setup empire: couldn't create ship of design {ship_design} for fleet {fleet_plan.name()}"
                )

    # starting resource stockpiles
    print("Player", player_name, ": add starting resource stockpiles")
    fo.empire_set_stockpile(empire, fo.resourceType.influence, 20.0)

    return True


def home_system_layout(home_systems, systems):  # noqa: C901
    """
    Home systems layout generation to place teams.
    Returns map from home system to neighbor home systems.
    """
    # for each system found nearest home systems
    # maybe multiple if home worlds placed on the same jump distnace
    system_hs = {}
    for system in systems:
        nearest_hs = set()
        nearest_dist = None
        for hs in home_systems:
            dist = fo.jump_distance(system, hs)
            if nearest_dist is None or nearest_dist > dist:
                nearest_dist = dist
                nearest_hs = {hs}
            elif nearest_dist == dist:
                nearest_hs.add(hs)
        system_hs[system] = nearest_hs

    # homeworld is connected to the other
    # if both are nearest for some system
    # if each of them is nearest for systems on the starline ends
    home_system_connections = {}
    for system, connected_home_systems in system_hs.items():
        if len(connected_home_systems) >= 2:
            for hs1, hs2 in unique_product(connected_home_systems, connected_home_systems):
                home_system_connections.setdefault(hs1, set()).add(hs2)
                home_system_connections.setdefault(hs2, set()).add(hs1)

    for system, connected_home_systems in system_hs.items():
        adj_systems = fo.systems_within_jumps_unordered(1, [system])
        for system2 in adj_systems:
            connected_home_systems2 = system_hs.get(system2, set())
            for hs1, hs2 in unique_product(connected_home_systems, connected_home_systems2):
                home_system_connections.setdefault(hs1, set()).add(hs2)
                home_system_connections.setdefault(hs2, set()).add(hs1)
    return home_system_connections



from typing_extensions import TypeAlias

from focs._types import _EffectGroup, _FloatParam

class _Item: ...
class _ItemType: ...

UnlockPolicy = _ItemType()
UnlockShipPart = _ItemType()
UnlockBuilding = _ItemType()
UnlockShipHull = _ItemType()

def Item(type: _ItemType, name: str) -> _Item: ...

_Color: TypeAlias = tuple[int, int, int] | tuple[int, int, int, int]

class _Category: ...

def Category(
    *,
    name: str,
    graphic: str,
    colour: _Color,
) -> _Category: ...
def Tech(
    name: str,
    description: str,
    short_description: str,
    category: str,
    researchcost: _FloatParam,
    researchturns: int,
    tags: list[str],
    prerequisites: list[str] = [],
    effectsgroups: list[_EffectGroup] = [],
    graphic: str = "",
    researchable=True,
    unlock=_Item(),
): ...



import freeorion as fo
from random import choice, sample, uniform

from util import report_error


def generate_fields(systems):
    """
    Generates stationary fields in some randomly chosen empty no star systems.
    """
    # filter out all empty no star systems
    candidates = [s for s in systems if (fo.sys_get_star_type(s) == fo.starType.noStar) and (not fo.sys_get_planets(s))]
    # make sure we have at least one empty no star system, otherwise return without creating any fields
    if not candidates:
        print("...no empty no star systems found, no fields created")
        return
    # pick 10-15% of all empty no star systems to create stationary fields in them, but at least one
    accepted = sample(candidates, max(int(len(candidates) * uniform(0.1, 0.15)), 1))
    for system in accepted:
        # randomly pick a field type
        field_type = choice(["FLD_NEBULA_1", "FLD_NEBULA_2", "FLD_NEBULA_3"])
        # and create the field
        if fo.create_field_in_system(field_type, uniform(40, 120), system) == fo.invalid_object():
            # create field failed, report an error
            report_error("Python generate_fields: create field %s in system %d failed" % (field_type, system))
    print("...fields created in %d systems out of %d empty no star systems" % (len(accepted), len(candidates)))



from focs._effects import AddSpecial, EffectsGroup, IsSource, NoStar, Star, Turn
from focs._species import *
from species.species_macros.empire_opinions import COMMON_OPINION_EFFECTS
from species.species_macros.env import NARROW_EP, TUNDRA_NARROW_EP
from species.species_macros.focus import (
    HAS_ADVANCED_FOCI,
    HAS_INDUSTRY_FOCUS,
    HAS_RESEARCH_FOCUS,
)
from species.species_macros.happiness import AVERAGE_HAPPINESS
from species.species_macros.industry import GOOD_INDUSTRY
from species.species_macros.influence import NO_INFLUENCE
from species.species_macros.native_fortification import DEFAULT_NATIVE_DEFENSE
from species.species_macros.population import BAD_POPULATION
from species.species_macros.research import VERY_BAD_RESEARCH
from species.species_macros.shields import STANDARD_SHIP_SHIELDS
from species.species_macros.stockpile import GOOD_STOCKPILE
from species.species_macros.supply import AVERAGE_SUPPLY
from species.species_macros.troops import BAD_DEFENSE_TROOPS, BAD_OFFENSE_TROOPS

Species(
    name="SP_SLEEPERS",
    description="SP_SLEEPERS_DESC",
    gameplay_description="SP_SLEEPERS_GAMEPLAY_DESC",
    native=True,
    tags=[
        "ORGANIC",
        "BAD_POPULATION",
        "VERY_BAD_RESEARCH",
        "NO_INFLUENCE",
        "BAD_OFFENSE_TROOPS",
        "GOOD_INDUSTRY",
        "GOOD_STOCKPILE",
        "PEDIA_ORGANIC_SPECIES_CLASS",
    ],
    foci=[
        HAS_INDUSTRY_FOCUS,
        HAS_RESEARCH_FOCUS,
        *HAS_ADVANCED_FOCI,
    ],
    defaultfocus="FOCUS_INDUSTRY",
    likes=[  # not much really, because they are boring,
        "SHIMMER_SILK_SPECIAL",  # nice and comfy for the long sleep and if u have to get out to write the name in the snow u can c where u lay when u get back cuz is shimmering
        "BLD_SOL_ORB_GEN",  # bright and shiny and we likes the industrial benefits
        "PLC_DREAM_RECURSION",  # dreaming sounds nice. good night. sweet dream recursions.
        "PLC_INDUSTRIALISM",  # automation. also nice. just dont expect us to work ourselves..
        "FOCUS_RESEARCH",  # i.e. dreaming
    ],
    dislikes=[  # not much really, because that would make them interesting,
        "SP_NIGHTSIDERS",  # who doesnt?
        "FOCUS_INDUSTRY",  # just cuz we're good at it dont meen we like it. go away. let us sleep. we're not worth u'r trouble anyway.
        "MIMETIC_ALLOY_SPECIAL",  # too obtrusive
    ],
    # hibernate. dreaming. somnambulism. automation.
    effectsgroups=[
        *GOOD_INDUSTRY,
        *VERY_BAD_RESEARCH,  # because they like to sleep a lot and when not then sit around and do nothing on their eccentric worlds ..
        *NO_INFLUENCE,
        *GOOD_STOCKPILE,
        *BAD_POPULATION,
        *AVERAGE_HAPPINESS,
        COMMON_OPINION_EFFECTS("SP_SLEEPERS"),
        *AVERAGE_SUPPLY,
        *BAD_OFFENSE_TROOPS,
        *BAD_DEFENSE_TROOPS,
        # not for description,
        *DEFAULT_NATIVE_DEFENSE,
        *STANDARD_SHIP_SHIELDS,
        NARROW_EP,
        # they live on eccentric worlds,
        EffectsGroup(
            scope=IsSource,
            activation=Turn(high=0) & ~Star(type=[NoStar]),
            effects=[AddSpecial(name="ECCENTRIC_ORBIT_SPECIAL")],
        ),
    ],
    environments=TUNDRA_NARROW_EP,
    graphic="icons/species/quadruped-05.png",
)



from buildings.buildings import SPECIES_LIKES_OR_DISLIKES_BUILDING_STABILITY_EFFECTS
from focs._effects import (
    ContainedBy,
    Contains,
    EffectsGroup,
    Happiness,
    IsBuilding,
    NamedReal,
    Object,
    OwnedBy,
    Planet,
    Population,
    SetTargetResearch,
    Source,
    TargetPopulation,
    Value,
)
from macros.base_prod import BUILDING_COST_MULTIPLIER
from macros.enqueue import ENQUEUE_BUILD_ONE_PER_PLANET
from macros.priorities import TARGET_AFTER_2ND_SCALING_PRIORITY

try:
    from focs._buildings import *
except ModuleNotFoundError:
    pass

BuildingType(  # type: ignore[reportUnboundVariable]
    name="BLD_AUTO_HISTORY_ANALYSER",
    description="BLD_AUTO_HISTORY_ANALYSER_DESC",
    buildcost=100 * BUILDING_COST_MULTIPLIER,
    buildtime=5,
    location=(
        Planet()
        & Contains(IsBuilding(name=["BLD_CULTURE_ARCHIVES"]))
        & ~Contains(IsBuilding(name=["BLD_AUTO_HISTORY_ANALYSER"]))
        & OwnedBy(empire=Source.Owner)
        & TargetPopulation(low=1)
    ),
    enqueuelocation=ENQUEUE_BUILD_ONE_PER_PLANET,
    effectsgroups=[
        *SPECIES_LIKES_OR_DISLIKES_BUILDING_STABILITY_EFFECTS,
        EffectsGroup(
            scope=(
                Object(id=Source.PlanetID)
                & Planet()
                & ~Population(high=0)
                & OwnedBy(empire=Source.Owner)
                & Happiness(low=0)
            ),
            activation=(ContainedBy(Object(id=Source.PlanetID) & Contains(IsBuilding(name=["BLD_CULTURE_ARCHIVES"])))),
            priority=TARGET_AFTER_2ND_SCALING_PRIORITY,
            effects=[
                SetTargetResearch(
                    value=Value + NamedReal(name="BLD_AUTO_HISTORY_ANALYSER_TARGET_RESEARCH_FLAT", value=5)
                )
            ],
        ),
    ],
    icon="icons/building/science1.png",
)



from unittest.mock import Mock

from common.statistic_interface import StatKey
from freeorion_tools.statistics import Stats


def test_stat_key():
    writer = Mock()
    stats = Stats(writer=writer)
    stats.write(StatKey.SHIP_CONT, 11)
    writer.assert_called_once_with("##ShipCount:11")



from focs._effects import (
    Contains,
    EffectsGroup,
    EmpireHasAdoptedPolicy,
    GiveEmpireTech,
    Good,
    HasSpecies,
    HasTag,
    IsBuilding,
    IsSource,
    IsTarget,
    JumpsBetween,
    LocalCandidate,
    MaxOf,
    NamedIntegerLookup,
    NamedReal,
    NamedRealLookup,
    Number,
    OwnedBy,
    Planet,
    RootCandidate,
    SetTargetHappiness,
    SetTargetPopulation,
    Source,
    Statistic,
    StatisticCount,
    Sum,
    Target,
    Turn,
    Value,
    VisibleToEmpire,
    WithinStarlaneJumps,
)
from macros.priorities import TARGET_AFTER_SCALING_PRIORITY, TARGET_POPULATION_LAST_BEFORE_OVERRIDE_PRIORITY


def _condition_other_species_nearby():
    return (
        Planet()
        & HasSpecies()
        & ~HasSpecies(name=[Source.Species])
        & ~HasTag(name="LOW_BRAINPOWER")
        & ~Contains(IsBuilding(name=["BLD_CONC_CAMP"]))
        & VisibleToEmpire(empire=Source.Owner)
        & (~OwnedBy(empire=Source.Owner) | ~EmpireHasAdoptedPolicy(empire=Source.Owner, name="PLC_RACIAL_PURITY"))
        & WithinStarlaneJumps(jumps=NamedIntegerLookup(name="XENOPHOBIC_MAX_JUMPS"), condition=IsSource)
    )


def _xenophobic_selfsustaining_qualifying_planet_count():
    return StatisticCount(
        int,
        condition=Planet()
        & HasSpecies()
        & ~(HasSpecies(name=[Source.Species]) | HasTag(name="LOW_BRAINPOWER"))
        & WithinStarlaneJumps(jumps=NamedIntegerLookup(name="XENOPHOBIC_MAX_JUMPS"), condition=IsSource),
    )


XENOPHOBIC_SELF = [
    # Give the concentration camp tech at game start
    EffectsGroup(
        scope=IsSource,
        activation=Turn(high=0),
        effects=GiveEmpireTech(name="CON_CONC_CAMP"),
    ),
    # Xenophobic Frenzy: stability malus for xenophobic species nearby other species
    EffectsGroup(
        scope=IsSource,
        activation=Number(low=1, condition=_condition_other_species_nearby()),
        stackinggroup="XENOPHOBIC_LABEL_SELF",
        accountinglabel="XENOPHOBIC_LABEL_SELF",
        priority=TARGET_AFTER_SCALING_PRIORITY,
        effects=(
            SetTargetHappiness(
                value=Value
                + NamedReal(name="XENOPHOBIC_TARGET_STABILITY_PERJUMP", value=-0.5)
                * Statistic(
                    float,
                    Sum,
                    value=1
                    + NamedIntegerLookup(name="XENOPHOBIC_MAX_JUMPS")
                    - JumpsBetween(Target.ID, LocalCandidate.ID),
                    condition=Planet()
                    & VisibleToEmpire(empire=Source.Owner)
                    & HasSpecies()
                    & ~HasSpecies(name=[Source.Species])
                    & ~HasTag(name="LOW_BRAINPOWER")
                    & ~Contains(IsBuilding(name=["BLD_CONC_CAMP"]))
                    & (
                        ~OwnedBy(empire=RootCandidate.Owner)
                        | ~EmpireHasAdoptedPolicy(empire=Source.Owner, name="PLC_RACIAL_PURITY")
                    )
                    & WithinStarlaneJumps(jumps=NamedIntegerLookup(name="XENOPHOBIC_MAX_JUMPS"), condition=IsSource),
                )
            )
        ),
    ),
    # Pop malus for self-sustaining xenophobic species based on the number of nearby planets with other species
    EffectsGroup(
        scope=IsSource,
        activation=Planet(environment=[Good])
        & HasTag(name="SELF_SUSTAINING")
        & (0 < _xenophobic_selfsustaining_qualifying_planet_count()),
        stackinggroup="XENOPHOBIC_POP_SELF",
        accountinglabel="XENOPHOBIC_LABEL_SELF",
        priority=TARGET_POPULATION_LAST_BEFORE_OVERRIDE_PRIORITY,
        effects=SetTargetPopulation(
            value=Value
            + MaxOf(
                float,
                -3 * Target.HabitableSize,  # Cap malus at the self-sustaining bonus
                Target.HabitableSize
                * NamedReal(name="XENOPHOBIC_TARGET_POPULATION_PERJUMP", value=-0.2)
                * Statistic(
                    float,
                    Sum,
                    value=1
                    + NamedIntegerLookup(name="XENOPHOBIC_MAX_JUMPS")
                    - JumpsBetween(Target.ID, LocalCandidate.ID),
                    condition=Planet()
                    & HasSpecies()
                    & ~HasSpecies(name=[Source.Species])
                    & ~HasTag(name="LOW_BRAINPOWER")
                    & WithinStarlaneJumps(jumps=NamedIntegerLookup(name="XENOPHOBIC_MAX_JUMPS"), condition=IsSource),
                ),
            )
        ),
    ),
]


# single argument should be the name of the species capitalized
def XENOPHOBIC_OTHER(name):
    # Xenophobic Harassment: stability malus to non-exobot species nearby different, xenophobic species
    return EffectsGroup(
        scope=_condition_other_species_nearby(),
        activation=Planet(),
        stackinggroup=f"XENOPHOBIC_LABEL_{name}_OTHER",
        accountinglabel=f"XENOPHOBIC_LABEL_{name}_OTHER",
        priority=TARGET_AFTER_SCALING_PRIORITY,
        effects=SetTargetHappiness(
            value=Value
            + NamedRealLookup(name="XENOPHOBIC_TARGET_STABILITY_PERJUMP")
            * Statistic(
                float,
                Sum,
                value=1 + NamedIntegerLookup(name="XENOPHOBIC_MAX_JUMPS") - JumpsBetween(Target.ID, LocalCandidate.ID),
                condition=Planet()
                & HasSpecies(name=[Source.Species])
                & (
                    ~OwnedBy(empire=Target.Owner)
                    | ~EmpireHasAdoptedPolicy(empire=LocalCandidate.Owner, name="PLC_RACIAL_PURITY")
                )
                & WithinStarlaneJumps(jumps=NamedIntegerLookup(name="XENOPHOBIC_MAX_JUMPS"), condition=IsTarget),
            )
        ),
    )



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="PRO_INDUSTRY_CENTER_I",
    description="PRO_INDUSTRY_CENTER_I_DESC",
    short_description="BUILDING_UNLOCK_SHORT_DESC",
    category="PRODUCTION_CATEGORY",
    researchcost=60 * TECH_COST_MULTIPLIER,
    researchturns=4,
    tags=["PEDIA_PRODUCTION_CATEGORY"],
    prerequisites=["PRO_ROBOTIC_PROD"],
    unlock=[
        Item(type=UnlockBuilding, name="BLD_INDUSTRY_CENTER"),
        Item(type=UnlockPolicy, name="PLC_INDUSTRIALISM"),
    ],
    graphic="icons/tech/industrial_centre_i.png",
)



from focs._tech import *

Tech(
    name="TECH_COL_MISIORLA",
    description="TECH_COL_MISIORLA_DESC",
    short_description="EXTINCT_UNLOCK_SHORT_DESC",
    category="SPY_CATEGORY",
    researchcost=1,
    researchturns=1,
    researchable=False,
    tags=["PEDIA_EXTINCT_SPECIES_TECHS", "SP_MISIORLA", "CTRL_EXTINCT"],
    graphic="icons/species/misiorla.png",
)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="SHP_QUANT_ENRG_MAG",
    description="SHP_QUANT_ENRG_MAG_DESC",
    short_description="SHIP_HULL_UNLOCK_SHORT_DESC",
    category="SHIP_HULLS_CATEGORY",
    researchcost=600 * TECH_COST_MULTIPLIER,
    researchturns=5,
    tags=["PEDIA_ENERGY_HULL_TECHS"],
    prerequisites=["LRN_GRAVITONICS", "SHP_ENRG_FRIGATE"],
    unlock=[
        Item(type=UnlockShipHull, name="SH_QUANTUM_ENERGY"),
        Item(type=UnlockBuilding, name="BLD_SHIPYARD_ENRG_SOLAR"),
    ],
    graphic="icons/ship_hulls/quantum_energy_hull_small.png",
)



import freeOrionAIInterface as fo
from collections.abc import Mapping
from logging import error, warning

from AIDependencies import INVALID_ID
from common.fo_typing import SystemId
from freeorion_tools.caching import cache_for_current_turn


@cache_for_current_turn
def _get_system_supply_map() -> dict[SystemId, int]:
    return fo.getEmpire().supplyProjections()  # map from system_id to supply


def get_system_supply(sys_id: SystemId) -> int:
    """
    Get the supply level of a system.

    Negative values indicate jumps away from supply. Return -99 if system is not connected.
    """
    far_away_from_supply = -99

    retval = _get_system_supply_map().get(sys_id, None)
    if retval is None:
        # This is only expected to happen if a system has no path to any supplied system.
        # As the current code should not allow such queries, this is logged as warning.
        # If future code breaks this assumption, feel free to adjust logging.
        warning("Queried supply value of a system not mapped in empire.supplyProjections(): %d" % sys_id)
        return far_away_from_supply
    return retval


@cache_for_current_turn
def _get_systems_map_by_jumps_to_supply() -> Mapping[int, tuple[SystemId]]:
    systems_by_jumps_to_supply = {}

    for sys_id, supply_val in _get_system_supply_map().items():
        systems_by_jumps_to_supply.setdefault(min(0, supply_val), []).append(sys_id)
    return {key: tuple(value) for key, value in systems_by_jumps_to_supply.items()}


def get_systems_by_supply_tier(supply_tier: int) -> tuple[SystemId]:
    """
    Get systems with supply tier.

    The current implementation does not distinguish between positive supply levels and caps at 0.
    Negative values indicate jumps away from supply.
    """
    if supply_tier > 0:
        warning(
            "The current implementation does not distinguish between positive supply levels. "
            "Interpreting the query as supply_tier=0 (indicating system in supply)."
        )
        supply_tier = 0
    return _get_systems_map_by_jumps_to_supply().get(supply_tier, tuple())


@cache_for_current_turn
def _get_enemy_supply_distance_map() -> Mapping[int, int]:
    enemies = [fo.getEmpire(_id) for _id in fo.allEmpireIDs() if _id != fo.empireID()]
    distance_to_enemy_supply = {}

    for enemy in enemies:
        if enemy is None:
            error("Got None for enemy empire!")
            continue

        for sys_id, supply_val in enemy.supplyProjections().items():
            distance_to_enemy_supply[sys_id] = min(distance_to_enemy_supply.get(sys_id, 999), -supply_val)

    return distance_to_enemy_supply


def get_distance_to_enemy_supply(sys_id: int) -> int:
    return _get_enemy_supply_distance_map().get(sys_id, 999)


@cache_for_current_turn
def _system_to_supply_group() -> Mapping[SystemId, int]:
    """
    Create a mapping of SystemIds to supply_group_id.
    Used buy supply_connected, get_supply_group_id and get_supply_group.
    """
    result = {}
    for num, group in enumerate(fo.getEmpire().resourceSupplyGroups, start=1):
        result.update((sys_id, num) for sys_id in group)
    return result


def supply_connected(system_id1: SystemId, system_id2: SystemId) -> bool:
    """Return whether the two systems are connected by this empire's supply chains."""
    mapping = _system_to_supply_group()
    # Different defaults to get False when neither system is part of any group
    return mapping.get(system_id1, -1) == mapping.get(system_id2, -2)


def get_supply_group_id(system_id: SystemId) -> int:
    """
    Return a number identifying a supply group, INVALID_ID if none.
    These number have no meaning beyond the current turn, groups may be completely different next turn.
    The number can be used to check if a set of system belongs to the same or multiple groups or as input to
    get_supply_group
    """
    return _system_to_supply_group().get(system_id, INVALID_ID)


def get_supply_group(group_id: int) -> fo.IntSet:
    """
    Returns a Set of systems in the given group. Note that group_ids are only valid for the current turn,
    so the argument must have been returned by get_supply_group_id in the same turn.
    """
    # There is no way to directly get the nth element, but it should iterate in the same order every time,
    # see _system_to_supply_group.
    for num, group in enumerate(fo.getEmpire().resourceSupplyGroups, start=1):
        if num == group_id:
            return group
    raise (ValueError(f"invalid supply group_id: {group_id}"))



"""
Config for tests inside this directory.
https://docs.pytest.org/en/2.7.3/plugins.html?highlight=re#working-with-plugins-and-conftest-files
"""

import os
import sys
from unittest.mock import MagicMock

this_dir = os.path.dirname(__file__)
sys.path.append(os.path.join(this_dir, "..", "..", "python", "AI/"))
sys.path.append(os.path.join(this_dir, "..", "..", "python"))


class aggression:
    beginner = 0
    turtle = 1
    cautious = 2
    typical = 3
    aggressive = 4
    maniacal = 5


def userString(x):
    return f"UserString {x}"


fo = MagicMock(name="freeOrionAIInterface")
fo.aggression = aggression
fo.userString = userString

sys.modules["freeOrionAIInterface"] = fo



from focs._effects import (
    AddSpecial,
    EffectsGroup,
    GenerateSitRepMessage,
    GiveEmpireTech,
    HasSpecial,
    IsSource,
    OwnerHasTech,
    Planet,
    RemoveSpecial,
    SetPopulation,
    SetTargetPopulation,
    Source,
    Target,
    Turn,
    Uninhabitable,
)
from focs._species import *
from macros.priorities import POPULATION_OVERRIDE_PRIORITY, TARGET_POPULATION_OVERRIDE_PRIORITY
from species.species_macros.advanced_focus import ADVANCED_FOCUS_EFFECTS
from species.species_macros.env import DESERT_STANDARD_EP
from species.species_macros.focus import (
    HAS_ADVANCED_FOCI,
    HAS_INFLUENCE_FOCUS,
    HAS_RESEARCH_FOCUS,
)
from species.species_macros.general import FOCUS_CHANGE_PENALTY, STANDARD_CONSTRUCTION, STANDARD_METER_GROWTH
from species.species_macros.happiness import AVERAGE_HAPPINESS
from species.species_macros.industry import NO_INDUSTRY
from species.species_macros.influence import VERY_BAD_INFLUENCE
from species.species_macros.planet_defense import AVERAGE_PLANETARY_DEFENSE
from species.species_macros.planet_shields import AVERAGE_PLANETARY_SHIELDS
from species.species_macros.research import GOOD_RESEARCH
from species.species_macros.shields import STANDARD_SHIP_SHIELDS
from species.species_macros.stealth import GOOD_STEALTH
from species.species_macros.stockpile import AVERAGE_STOCKPILE
from species.species_macros.supply import AVERAGE_SUPPLY
from species.species_macros.troops import AVERAGE_DEFENSE_TROOPS

Species(
    name="SP_LEMBALALAM",
    description="SP_LEMBALALAM_DESC",
    gameplay_description="SP_LEMBALALAM_GAMEPLAY_DESC",
    native=True,
    tags=[
        "ORGANIC",
        "NO_INDUSTRY",
        "GOOD_RESEARCH",
        "VERY_BAD_INFLUENCE",
        "AVERAGE_SUPPLY",
        "GOOD_STEALTH",
        "PEDIA_ORGANIC_SPECIES_CLASS",
    ],
    foci=[
        # HAS_INDUSTRY_FOCUS
        HAS_RESEARCH_FOCUS,
        HAS_INFLUENCE_FOCUS,
        # HAS_GROWTH_FOCUS
        *HAS_ADVANCED_FOCI,
    ],
    defaultfocus="FOCUS_RESEARCH",
    likes=[
        "FOCUS_RESEARCH",
        "SPARK_FOSSILS_SPECIAL",
        "SUCCULENT_BARNACLES_SPECIAL",
        "ANCIENT_RUINS_DEPLETED_SPECIAL",
        "PLC_BUREAUCRACY",
        "PLC_AUGMENTATION",
        "PLC_CHECKPOINTS",
        "PLC_EXOBOT_PRODUCTIVITY",
        "PLC_ISOLATION",
        "PLC_NO_GROWTH",
    ],
    dislikes=[
        "SUPERCONDUCTOR_SPECIAL",
        "SPICE_SPECIAL",
        "PROBIOTIC_SPECIAL",
        "POSITRONIUM_SPECIAL",
        "MONOPOLE_SPECIAL",
        "MINERALS_SPECIAL",
        "PLC_MARTIAL_LAW",
        "PLC_SYSTEM_INFRA",
        "PLC_INTERSTELLAR_INFRA",
        "PLC_METROPOLES",
        "PLC_POPULATION",
        "SP_NIGHTSIDERS",
    ],
    effectsgroups=[
        NO_INDUSTRY,
        *GOOD_RESEARCH,
        *VERY_BAD_INFLUENCE,
        *AVERAGE_STOCKPILE,
        *AVERAGE_HAPPINESS,
        *AVERAGE_SUPPLY,
        *AVERAGE_DEFENSE_TROOPS,
        *GOOD_STEALTH,
        # other species get these via BASIC_POPULATION,
        *FOCUS_CHANGE_PENALTY,
        *ADVANCED_FOCUS_EFFECTS,
        *STANDARD_CONSTRUCTION,
        *STANDARD_METER_GROWTH,
        # not for description,
        *AVERAGE_PLANETARY_SHIELDS,
        *AVERAGE_PLANETARY_DEFENSE,
        *STANDARD_SHIP_SHIELDS,
        # Start the game with Volcanic Ash stealth special,
        EffectsGroup(
            scope=IsSource,
            activation=Turn(high=0),
            # stackinggroup = "GAME_START_MOD_STACK",
            effects=[
                AddSpecial(name="VOLCANIC_ASH_SLAVE_SPECIAL"),
                AddSpecial(name="VOLCANIC_ASH_MASTER_SPECIAL"),
            ],
        ),
        # Always start the game as High Tech natives,
        EffectsGroup(
            scope=IsSource & HasSpecial(name="MODERATE_TECH_NATIVES_SPECIAL"),
            activation=Turn(high=0),
            effects=RemoveSpecial(name="MODERATE_TECH_NATIVES_SPECIAL"),
        ),
        EffectsGroup(
            scope=IsSource & ~HasSpecial(name="HIGH_TECH_NATIVES_SPECIAL"),
            activation=Turn(high=0),
            # stackinggroup = "GAME_START_MOD_STACK",
            effects=AddSpecial(name="HIGH_TECH_NATIVES_SPECIAL"),
        ),
        # Max Population is fixed at 5,
        EffectsGroup(
            description="FIXED_LOW_POPULATION_DESC",
            scope=IsSource,
            activation=Planet() & ~Planet(environment=[Uninhabitable]),
            priority=TARGET_POPULATION_OVERRIDE_PRIORITY,
            effects=SetTargetPopulation(value=5, accountinglabel="IMMORTAL_LABEL"),
        ),
        # set population to max,
        EffectsGroup(
            scope=IsSource & Planet(),
            activation=Turn(high=1),
            priority=POPULATION_OVERRIDE_PRIORITY,
            effects=SetPopulation(value=Target.TargetPopulation),
        ),
        # Give techs to empire,
        EffectsGroup(
            scope=IsSource,
            activation=~OwnerHasTech(name="SPY_STEALTH_1"),
            stackinggroup="LEMBALALAM_TECH_UNLOCK",
            effects=[
                GiveEmpireTech(name="SPY_STEALTH_1", empire=Target.Owner),
                GenerateSitRepMessage(
                    message="EFFECT_NATIVES_TECH",
                    label="EFFECT_NATIVES_TECH_LABEL",
                    icon="icons/species/other-01.png",
                    parameters={
                        "planet": Source.ID,
                        "species": "SP_LEMBALALAM",
                        "tech": "SPY_STEALTH_1",
                    },
                    empire=Source.Owner,
                ),
            ],
        ),
        EffectsGroup(
            scope=IsSource,
            activation=~OwnerHasTech(name="SPY_STEALTH_2") & OwnerHasTech(name="SPY_STEALTH_1"),
            stackinggroup="LEMBALALAM_TECH_UNLOCK",
            effects=[
                GiveEmpireTech(name="SPY_STEALTH_2", empire=Target.Owner),
                GenerateSitRepMessage(
                    message="EFFECT_NATIVES_TECH",
                    label="EFFECT_NATIVES_TECH_LABEL",
                    icon="icons/species/other-01.png",
                    parameters={
                        "planet": Source.ID,
                        "species": "SP_LEMBALALAM",
                        "tech": "SPY_STEALTH_2",
                    },
                    empire=Source.Owner,
                ),
            ],
        ),
        EffectsGroup(
            scope=IsSource,
            activation=~OwnerHasTech(name="GRO_LIFECYCLE_MAN"),
            stackinggroup="LEMBALALAM_TECH_UNLOCK",
            effects=[
                GiveEmpireTech(name="GRO_LIFECYCLE_MAN", empire=Target.Owner),
                GenerateSitRepMessage(
                    message="EFFECT_NATIVES_TECH",
                    label="EFFECT_NATIVES_TECH_LABEL",
                    icon="icons/species/other-01.png",
                    parameters={
                        "planet": Source.ID,
                        "species": "SP_LEMBALALAM",
                        "tech": "GRO_LIFECYCLE_MAN",
                    },
                    empire=Source.Owner,
                ),
            ],
        ),
    ],
    environments=DESERT_STANDARD_EP,
    graphic="icons/species/other-01.png",
)



from focs._species import *
from species.species_macros.empire_opinions import COMMON_OPINION_EFFECTS
from species.species_macros.env import SWAMP_STANDARD_EP
from species.species_macros.focus import (
    HAS_ADVANCED_FOCI,
    HAS_GROWTH_FOCUS,
    HAS_INDUSTRY_FOCUS,
    HAS_INFLUENCE_FOCUS,
    HAS_RESEARCH_FOCUS,
)
from species.species_macros.happiness import AVERAGE_HAPPINESS
from species.species_macros.industry import BAD_INDUSTRY
from species.species_macros.influence import AVERAGE_INFLUENCE
from species.species_macros.native_fortification import DEFAULT_NATIVE_DEFENSE
from species.species_macros.population import GOOD_POPULATION
from species.species_macros.research import AVERAGE_RESEARCH
from species.species_macros.shields import STANDARD_SHIP_SHIELDS
from species.species_macros.stealth import GREAT_STEALTH
from species.species_macros.stockpile import AVERAGE_STOCKPILE
from species.species_macros.supply import AVERAGE_SUPPLY
from species.species_macros.troops import NO_DEFENSE_TROOPS, NO_OFFENSE_TROOPS

Species(
    name="SP_SETINON",
    description="SP_SETINON_DESC",
    gameplay_description="SP_SETINON_GAMEPLAY_DESC",
    native=True,
    can_produce_ships=True,
    can_colonize=True,
    tags=[
        "ORGANIC",
        "BAD_INDUSTRY",
        "GOOD_POPULATION",
        "AVERAGE_SUPPLY",
        "GREAT_STEALTH",
        "NO_OFFENSE_TROOPS",
        "PEDIA_ORGANIC_SPECIES_CLASS",
    ],
    foci=[
        HAS_INDUSTRY_FOCUS,
        HAS_RESEARCH_FOCUS,
        HAS_INFLUENCE_FOCUS,
        HAS_GROWTH_FOCUS,
        *HAS_ADVANCED_FOCI,
    ],
    defaultfocus="FOCUS_PROTECTION",
    likes=[
        "FOCUS_PROTECTION",
        "FRACTAL_GEODES_SPECIAL",
        "SUCCULENT_BARNACLES_SPECIAL",
        "WORLDTREE_SPECIAL",
        "FORTRESS_SPECIAL",
        "CRYSTALS_SPECIAL",
        "COMPUTRONIUM_SPECIAL",
        "PLC_MARTIAL_LAW",
        "PLC_CONFEDERATION",
        "PLC_ARTISAN_WORKSHOPS",
    ],
    dislikes=[
        "BLD_SHIPYARD_ENRG_SOLAR",
        "BLD_SCANNING_FACILITY",
        "BLD_INDUSTRY_CENTER",
        "BLD_INTERSPECIES_ACADEMY",
        "BLD_STOCKPILING_CENTER",
        "BLD_MEGALITH",
        "BLD_SPACE_ELEVATOR",
        "SUPERCONDUCTOR_SPECIAL",
        "SPICE_SPECIAL",
        "PROBIOTIC_SPECIAL",
        "POSITRONIUM_SPECIAL",
        "PLC_CENTRALIZATION",
        "PLC_INDUSTRIALISM",
        "PLC_CONFORMANCE",
        "PLC_DIVINE_AUTHORITY",
    ],
    effectsgroups=[
        *BAD_INDUSTRY,
        *AVERAGE_RESEARCH,
        *AVERAGE_INFLUENCE,
        *AVERAGE_STOCKPILE,
        *GOOD_POPULATION,
        *AVERAGE_HAPPINESS,
        COMMON_OPINION_EFFECTS("SP_SETINON"),
        *AVERAGE_SUPPLY,
        *NO_DEFENSE_TROOPS,
        *NO_OFFENSE_TROOPS,
        *GREAT_STEALTH,
        # not for description,
        *DEFAULT_NATIVE_DEFENSE,
        *STANDARD_SHIP_SHIELDS,
    ],
    environments=SWAMP_STANDARD_EP,
    graphic="icons/species/amorphous-02.png",
)



import logging
import sys
from code import InteractiveInterpreter
from io import StringIO

from freeorion_tools.chat_handler.shell_variable import ShellVariable


class DebugInterpreter:
    def __init__(self, shell_locals: list[ShellVariable]):
        self._shell_locals = shell_locals
        self._interpreter = InteractiveInterpreter()

    def eval(self, msg):
        old_stdout = sys.stdout
        old_stderr = sys.stderr

        sys.stdout = StringIO()
        sys.stderr = StringIO()
        if msg.endswith(";"):
            msg = msg.replace(";", "\n")
        handler = logging.StreamHandler(sys.stdout)
        logging.getLogger().addHandler(handler)
        self._interpreter.runsource(msg)
        logging.getLogger().removeHandler(handler)

        sys.stdout.seek(0)
        out = sys.stdout.read()
        sys.stderr.seek(0)
        err = sys.stderr.read()
        sys.stdout = old_stdout
        sys.stderr = old_stderr
        return out.strip("\n"), err.strip("\n")

    def set_locals(self):
        initial_code = []

        variables_description = []
        for shellLocal in self._shell_locals:
            initial_code.extend(shellLocal.get_evaluation_command())
            result = shellLocal.name, shellLocal.description
            variables_description.append(result)
        for line in initial_code:
            self.eval(line)

        return variables_description



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="SHP_CONT_SYMB",
    description="SHP_CONT_SYMB_DESC",
    short_description="SHIP_HULL_UNLOCK_SHORT_DESC",
    category="SHIP_HULLS_CATEGORY",
    researchcost=32 * TECH_COST_MULTIPLIER,
    researchturns=8,
    tags=["PEDIA_ORGANIC_HULL_TECHS"],
    prerequisites=["SHP_ORG_HULL"],
    unlock=[
        Item(type=UnlockShipHull, name="SH_SYMBIOTIC"),
        Item(type=UnlockBuilding, name="BLD_SHIPYARD_ORG_CELL_GRO_CHAMB"),
    ],
    graphic="icons/ship_hulls/symbiotic_hull_small.png",
)



import freeOrionAIInterface as fo
import math
from collections.abc import Sequence
from logging import debug, error, warning
from typing import Union

import AIDependencies
import MoveUtilsAI
from AIDependencies import INVALID_ID
from aistate_interface import get_aistate
from CombatRatingsAI import get_fleet_rating, get_ship_combat_stats, rating_needed
from common.fo_typing import PlanetId
from common.print_utils import Sequence as SequenceField
from common.print_utils import Table, Text
from EnumsAI import MissionType, ShipRoleType
from freeorion_tools import assertion_fails, combine_ratings
from ShipDesignAI import get_ship_part
from target import TargetFleet, TargetPlanet, TargetSystem
from universe.system_network import get_neighbors, get_shortest_distance


def stats_meet_reqs(stats: dict, requirements: dict) -> bool:
    """
    Check if (fleet) stats meet requirements.

    :param stats: Stats (of fleet)
    :param requirements: Requirements
    :return: True if requirements are met.
    """
    for key in requirements:
        if key not in stats:  # skip requirements not related to stats
            if key != "target_system":  # expected not to be in stats
                warning("Requirement %s not in stats", key)
            continue
        if stats.get(key, 0) < requirements[key]:
            return False
    return True


def count_troops_in_fleet(fleet_id: int) -> float:
    """
    Get the number of troops in the fleet.

    :param fleet_id: fleet to be queried
    :return: total troopCapacity of the fleet
    """
    universe = fo.getUniverse()
    fleet = universe.getFleet(fleet_id)
    if not fleet:
        return 0.0
    fleet_troop_capacity = 0.0
    for ship_id in fleet.shipIDs:
        ship = universe.getShip(ship_id)
        if ship:
            fleet_troop_capacity += ship.troopCapacity
    return fleet_troop_capacity


def get_targeted_planet_ids(planet_ids: Sequence[PlanetId], mission_type: MissionType) -> list[PlanetId]:
    """Find the planets that are targets of the specified mission type.

    :param planet_ids: planets to be queried
    :param mission_type:
    :return: Subset of *planet_ids* targeted by *mission_type*
    """
    selected_fleet_missions = get_aistate().get_fleet_missions_with_any_mission_types([mission_type])
    targeted_planets = []
    for planet_id in planet_ids:
        # add planets that are target of a mission
        for fleet_mission in selected_fleet_missions:
            ai_target = TargetPlanet(planet_id)
            if fleet_mission.has_target(mission_type, ai_target):
                targeted_planets.append(planet_id)
    return targeted_planets


# TODO: Avoid mutable arguments and use return values instead
# TODO: Use Dijkstra's algorithm instead of BFS to consider starlane length
def get_fleets_for_mission(  # noqa: C901
    target_stats: dict,
    min_stats: dict,
    cur_stats: dict,
    starting_system: int,
    fleet_pool_set: set[int],
    fleet_list: list[int],
    species: str = "",
    ensure_return: bool = False,
) -> list[int]:
    """Get fleets for a mission.

    Implements breadth-first search through systems starting at the **starting_sytem**.
    In each system, local fleets are checked if they are in the allowed **fleet_pool_set** and suitable for the mission.
    If so, they are added to the **fleet_list** and **cur_stats** is updated with the currently selected fleet summary.
    The search continues until the requirements defined in **target_stats** are met or there are no more systems/fleets.
    In that case, if the **min_stats** are covered, the **fleet_list** is returned anyway.
    Otherwise, an empty list is returned by the function, in which case the caller can make an evaluation of
    an emergency use of the found fleets in fleet_list; if not to be used they should be added back to the main pool.

    :param target_stats: stats the fleet should ideally meet
    :param min_stats: minimum stats the final fleet must meet to be accepted
    :param cur_stats: (**mutated**) stat summary of selected fleets
    :param starting_system: system_id where breadth-first-search is centered
    :param fleet_pool_set: (**mutated**) fleets allowed to be selected. Split fleed_ids are added, used ones removed.
    :param fleet_list: (**mutated**) fleets that are selected for the mission. Gets filled during the call.
    :param species: species for colonization mission
    :param bool ensure_return: If true, fleet must have sufficient fuel to return into supply after mission
    :return: List of selected fleet_ids or empty list if couldn't meet minimum requirements.
    """
    universe = fo.getUniverse()
    colonization_roles = (ShipRoleType.CIVILIAN_COLONISATION, ShipRoleType.BASE_COLONISATION)
    systems_enqueued = [starting_system]
    systems_visited = []
    # loop over systems in a breadth-first-search trying to find nearby suitable ships in fleet_pool_set
    aistate = get_aistate()
    while systems_enqueued and fleet_pool_set:
        this_system_id = systems_enqueued.pop(0)
        this_system_obj = TargetSystem(this_system_id)
        systems_visited.append(this_system_id)
        accessible_fleets = aistate.systemStatus.get(this_system_id, {}).get("myFleetsAccessible", [])
        fleets_here = [fid for fid in accessible_fleets if fid in fleet_pool_set]
        # loop over all fleets in the system, split them if possible and select suitable ships
        while fleets_here:
            fleet_id = fleets_here.pop(0)
            fleet = universe.getFleet(fleet_id)
            if not fleet:  # TODO should be checked before passed to the function
                fleet_pool_set.remove(fleet_id)
                continue
            # try splitting fleet
            if fleet.numShips > 1:
                debug("Splitting candidate fleet to get ships for mission.")
                new_fleets = split_fleet(fleet_id)
                fleet_pool_set.update(new_fleets)
                fleets_here.extend(new_fleets)

            if "target_system" in target_stats and not MoveUtilsAI.can_travel_to_system(
                fleet_id, this_system_obj, target_stats["target_system"], ensure_return=ensure_return
            ):
                continue

            # check species for colonization missions
            if species:
                for ship_id in fleet.shipIDs:
                    ship = universe.getShip(ship_id)
                    if (
                        ship
                        and aistate.get_ship_role(ship.design.id) in colonization_roles
                        and species == ship.speciesName
                    ):
                        break
                else:  # no suitable species found
                    continue
            # check troop capacity for invasion missions
            troop_capacity = 0
            if "troopCapacity" in target_stats:
                troop_capacity = count_troops_in_fleet(fleet_id)
                if troop_capacity <= 0:
                    continue

            # check if we need additional rating vs planets
            this_rating_vs_planets = 0
            if "ratingVsPlanets" in target_stats:
                this_rating_vs_planets = aistate.get_rating(fleet_id, against_planets=True)
                if this_rating_vs_planets <= 0 and cur_stats.get("rating", 0) >= target_stats.get("rating", 0):
                    # we already have enough general rating, so do not add any more warships useless against planets
                    continue

            # all checks passed, add ship to selected fleets and update the stats
            try:
                fleet_pool_set.remove(fleet_id)
            except KeyError:
                error("After having split a fleet, the original fleet apparently no longer exists.", exc_info=True)
                continue
            fleet_list.append(fleet_id)

            this_rating = aistate.get_rating(fleet_id)
            cur_stats["rating"] = combine_ratings(cur_stats.get("rating", 0), this_rating)
            if "ratingVsPlanets" in target_stats:
                cur_stats["ratingVsPlanets"] = combine_ratings(
                    cur_stats.get("ratingVsPlanets", 0), this_rating_vs_planets
                )
            if "troopCapacity" in target_stats:
                cur_stats["troopCapacity"] = cur_stats.get("troopCapacity", 0) + troop_capacity
            # if we already meet the requirements, we can stop looking for more ships
            if (sum(len(universe.getFleet(fid).shipIDs) for fid in fleet_list) >= 1) and stats_meet_reqs(
                cur_stats, target_stats
            ):
                return fleet_list

        # finished system without meeting requirements. Add neighboring systems to search queue.
        for neighbor_id in get_neighbors(this_system_id):
            if all(
                (
                    neighbor_id not in systems_visited,
                    neighbor_id not in systems_enqueued,
                    neighbor_id in aistate.exploredSystemIDs,
                )
            ):
                systems_enqueued.append(neighbor_id)
    # we ran out of systems or fleets to check but did not meet requirements yet.
    if stats_meet_reqs(cur_stats, min_stats) and any(universe.getFleet(fid).shipIDs for fid in fleet_list):
        return fleet_list
    else:
        return []


def split_fleet(fleet_id: int) -> list[int]:
    """Split a fleet into its ships.

    :param fleet_id: fleet to be split.
    :return: New fleets. Empty if couldn't split.
    """
    universe = fo.getUniverse()
    empire_id = fo.empireID()
    fleet = universe.getFleet(fleet_id)
    new_fleets = []

    if fleet is None:
        return []
    if not fleet.ownedBy(empire_id):
        return []

    if len(list(fleet.shipIDs)) <= 1:  # fleet with only one ship cannot be split
        return []
    ship_ids = list(fleet.shipIDs)
    aistate = get_aistate()
    for ship_id in ship_ids[1:]:
        new_fleet_id = split_ship_from_fleet(fleet_id, ship_id)
        new_fleets.append(new_fleet_id)

    aistate.get_fleet_role(fleet_id, force_new=True)
    aistate.update_fleet_rating(fleet_id)
    if new_fleets:
        aistate.ensure_have_fleet_missions(new_fleets)

    return new_fleets


def split_ship_from_fleet(fleet_id, ship_id) -> int:
    """Try to split a ship from the fleet, creating a new fleet.

    :return: ID of the newly created fleet or INVALID_ID if failed
    """
    universe = fo.getUniverse()
    fleet = universe.getFleet(fleet_id)
    if assertion_fails(fleet is not None):
        return INVALID_ID

    if assertion_fails(ship_id in fleet.shipIDs):
        return INVALID_ID

    if assertion_fails(fleet.numShips > 1, "Can't split last ship from fleet"):
        return INVALID_ID

    new_fleet_id = fo.issueNewFleetOrder("Fleet %4d" % ship_id, ship_id)
    if new_fleet_id != INVALID_ID:
        aistate = get_aistate()
        new_fleet = universe.getFleet(new_fleet_id)
        if not new_fleet:
            warning("Newly split fleet %d not available from universe" % new_fleet_id)
        debug("Successfully split ship %d from fleet %d into new fleet %d", ship_id, fleet_id, new_fleet_id)
        fo.issueRenameOrder(new_fleet_id, "Fleet %4d" % new_fleet_id)  # to ease review of debugging logs
        fo.issueAggressionOrder(new_fleet_id, True)
        aistate.update_fleet_rating(new_fleet_id)
        aistate.newlySplitFleets[new_fleet_id] = True
        # register the new fleets so AI logic is aware of them
        sys_status = aistate.systemStatus.setdefault(fleet.systemID, {})
        sys_status["myfleets"].append(new_fleet_id)
        sys_status["myFleetsAccessible"].append(new_fleet_id)
    else:
        if fleet.systemID == INVALID_ID:
            warning("Tried to split ship id (%d) from fleet %d when fleet is in starlane" % (ship_id, fleet_id))
        else:
            warning("Got no fleet ID back after trying to split ship id (%d) from fleet %d" % (ship_id, fleet_id))
    return new_fleet_id


def merge_fleet_a_into_b(fleet_a_id, fleet_b_id, leave_rating=0, need_rating=0, context=""):
    debug("Merging fleet %s into %s", TargetFleet(fleet_a_id), TargetFleet(fleet_b_id))
    universe = fo.getUniverse()
    fleet_a = universe.getFleet(fleet_a_id)
    fleet_b = universe.getFleet(fleet_b_id)
    if not fleet_a or not fleet_b:
        return 0
    system_id = fleet_a.systemID
    if fleet_b.systemID != system_id:
        return 0

    # TODO: Should this rate against specific enemy?
    remaining_rating = get_fleet_rating(fleet_a_id)
    transferred_rating = 0
    for ship_id in fleet_a.shipIDs:
        this_ship = universe.getShip(ship_id)
        if not this_ship:
            continue
        this_rating = get_ship_combat_stats(ship_id).get_rating()
        remaining_rating = rating_needed(remaining_rating, this_rating)
        if remaining_rating < leave_rating:  # merging this would leave old fleet under minimum rating, try other ships.
            continue
        transferred = fo.issueFleetTransferOrder(ship_id, fleet_b_id)
        if transferred:
            transferred_rating = combine_ratings(transferred_rating, this_rating)
        else:
            debug(
                "  *** transfer of ship %4d, formerly of fleet %4d, into fleet %4d failed; %s"
                % (ship_id, fleet_a_id, fleet_b_id, (" context is %s" % context) if context else "")
            )
        if need_rating != 0 and need_rating <= transferred_rating:
            break
    fleet_a = universe.getFleet(fleet_a_id)
    aistate = get_aistate()
    if not fleet_a or fleet_a.empty or fleet_a_id in universe.destroyedObjectIDs(fo.empireID()):
        aistate.delete_fleet_info(fleet_a_id)
    aistate.update_fleet_rating(fleet_b_id)


def fleet_has_ship_with_role(fleet_id, ship_role):
    """Returns True if a ship with shipRole is in the fleet."""
    universe = fo.getUniverse()
    fleet = universe.getFleet(fleet_id)

    if fleet is None:
        return False
    aistate = get_aistate()
    for ship_id in fleet.shipIDs:
        ship = universe.getShip(ship_id)
        if aistate.get_ship_role(ship.design.id) == ship_role:
            return True
    return False


def get_ship_id_with_role(fleet_id, ship_role, verbose=True):
    """Returns a ship with the specified role in the fleet."""

    if not fleet_has_ship_with_role(fleet_id, ship_role):
        if verbose:
            debug("No ship with role %s found." % ship_role)
        return None

    universe = fo.getUniverse()
    fleet = universe.getFleet(fleet_id)
    aistate = get_aistate()

    for ship_id in fleet.shipIDs:
        ship = universe.getShip(ship_id)
        if aistate.get_ship_role(ship.design.id) == ship_role:
            return ship_id


def get_empire_fleet_ids():
    """Returns all fleetIDs for current empire."""
    empire_id = fo.empireID()
    universe = fo.getUniverse()
    empire_fleet_ids = []
    destroyed_object_ids = universe.destroyedObjectIDs(empire_id)
    for fleet_id in set(list(universe.fleetIDs) + list(get_aistate().newlySplitFleets)):
        fleet = universe.getFleet(fleet_id)
        if fleet is None:
            continue
        if fleet.ownedBy(empire_id) and fleet_id not in destroyed_object_ids and not fleet.empty and fleet.shipIDs:
            empire_fleet_ids.append(fleet_id)
    return empire_fleet_ids


def get_empire_fleet_ids_by_role(fleet_role):
    """Returns a list with fleet_ids that have the specified role."""
    fleet_ids = get_empire_fleet_ids()
    fleet_ids_with_role = []
    aistate = get_aistate()
    for fleet_id in fleet_ids:
        if aistate.get_fleet_role(fleet_id) != fleet_role:
            continue
        fleet_ids_with_role.append(fleet_id)
    return fleet_ids_with_role


def extract_fleet_ids_without_mission_types(fleets_ids):
    """Extracts a list with fleetIDs that have no mission."""
    aistate = get_aistate()
    return [fleet_id for fleet_id in fleets_ids if not aistate.get_fleet_mission(fleet_id).type]


def assess_fleet_role(fleet_id):  # noqa: C901
    """
    Assesses ShipRoles represented in a fleet and
    returns a corresponding overall fleetRole (of type MissionType).
    """
    universe = fo.getUniverse()
    ship_roles = {}
    fleet = universe.getFleet(fleet_id)
    if not fleet:
        debug("couldn't get fleet with id " + str(fleet_id))
        return ShipRoleType.INVALID

    # count ship_roles
    aistate = get_aistate()
    for ship_id in fleet.shipIDs:
        ship = universe.getShip(ship_id)
        if ship.design:
            role = aistate.get_ship_role(ship.design.id)
        else:
            role = ShipRoleType.INVALID

        if role != ShipRoleType.INVALID:
            ship_roles[role] = ship_roles.get(role, 0) + 1
    # determine most common ship_role
    favourite_role = ShipRoleType.INVALID
    for ship_role in ship_roles:
        if ship_roles[ship_role] == max(ship_roles.values()):
            favourite_role = ship_role

    # assign fleet role
    if ShipRoleType.CIVILIAN_COLONISATION in ship_roles:
        selected_role = MissionType.COLONISATION
    elif ShipRoleType.BASE_COLONISATION in ship_roles:
        selected_role = MissionType.COLONISATION
    elif ShipRoleType.CIVILIAN_OUTPOST in ship_roles:
        selected_role = MissionType.OUTPOST
    elif ShipRoleType.BASE_OUTPOST in ship_roles:
        selected_role = MissionType.ORBITAL_OUTPOST
    elif ShipRoleType.BASE_INVASION in ship_roles:
        selected_role = MissionType.ORBITAL_INVASION
    elif ShipRoleType.BASE_DEFENSE in ship_roles:
        selected_role = MissionType.ORBITAL_DEFENSE
    elif ShipRoleType.MILITARY_INVASION in ship_roles:
        selected_role = MissionType.INVASION
    ####
    elif favourite_role == ShipRoleType.CIVILIAN_EXPLORATION:
        selected_role = MissionType.EXPLORATION
    elif favourite_role == ShipRoleType.MILITARY_ATTACK:
        selected_role = MissionType.MILITARY
    elif favourite_role == ShipRoleType.MILITARY:
        selected_role = MissionType.MILITARY
    else:
        selected_role = ShipRoleType.INVALID
    return selected_role


def assess_ship_design_role(design):  # noqa: C901
    parts = [get_ship_part(partname) for partname in design.parts if partname and get_ship_part(partname)]

    if any(p.partClass == fo.shipPartClass.colony and p.capacity == 0 for p in parts):
        if design.speed > 0:
            return ShipRoleType.CIVILIAN_OUTPOST
        else:
            return ShipRoleType.BASE_OUTPOST

    if any(p.partClass == fo.shipPartClass.colony and p.capacity > 0 for p in parts):
        if design.speed > 0:
            return ShipRoleType.CIVILIAN_COLONISATION
        else:
            return ShipRoleType.BASE_COLONISATION

    if any(p.partClass == fo.shipPartClass.troops for p in parts):
        if design.speed > 0:
            return ShipRoleType.MILITARY_INVASION
        else:
            return ShipRoleType.BASE_INVASION

    if design.speed == 0:
        if not parts or parts[0].partClass in (
            fo.shipPartClass.shortRange,
            fo.shipPartClass.fighterBay,
            fo.shipPartClass.fighterHangar,
            fo.shipPartClass.shields,
            fo.shipPartClass.armour,
        ):
            return ShipRoleType.BASE_DEFENSE
        else:
            return ShipRoleType.INVALID

    if design.isArmed or design.hasFighters:
        return ShipRoleType.MILITARY
    if any(p.partClass == fo.shipPartClass.detection for p in parts):
        return ShipRoleType.CIVILIAN_EXPLORATION
    else:  # if no suitable role found, use as (bad) scout as it still has inherent detection
        warning("Defaulting ship role to 'exploration' for ship with parts: %s", design.parts)
        return ShipRoleType.CIVILIAN_EXPLORATION


def generate_fleet_orders_for_fleet_missions():  # noqa: C901
    """Generates fleet orders from targets."""
    debug("Generating fleet orders")

    table = Table(Text("name"), SequenceField("values"), hide_header=True, table_name="Fleets by Role")

    # The following fleet lists are based on *Roles* -- Secure type missions are done by fleets with Military Roles
    table.add_row("Exploration", get_empire_fleet_ids_by_role(MissionType.EXPLORATION))
    table.add_row("Colonization", get_empire_fleet_ids_by_role(MissionType.COLONISATION))
    table.add_row("Outpost", get_empire_fleet_ids_by_role(MissionType.OUTPOST))
    table.add_row("Invasion", get_empire_fleet_ids_by_role(MissionType.INVASION))
    table.add_row("Military", get_empire_fleet_ids_by_role(MissionType.MILITARY))
    table.add_row("Orbital Defense", get_empire_fleet_ids_by_role(MissionType.ORBITAL_DEFENSE))
    table.add_row("Outpost Base", get_empire_fleet_ids_by_role(MissionType.ORBITAL_OUTPOST))
    table.add_row("Invasion Base", get_empire_fleet_ids_by_role(MissionType.ORBITAL_INVASION))
    table.add_row(
        "Securing",
        get_empire_fleet_ids_by_role(MissionType.SECURE),
        note="currently MissionType.MILITARY should be used instead of this role",
    )

    debug(table)

    aistate = get_aistate()
    if fo.currentTurn() < 50:
        debug("")
        debug("Explored systems:")
        _print_systems_and_supply(aistate.get_explored_system_ids())
        debug("Unexplored systems:")
        _print_systems_and_supply(aistate.get_unexplored_system_ids())
        debug("")

    exploration_fleet_missions = aistate.get_fleet_missions_with_any_mission_types([MissionType.EXPLORATION])
    if exploration_fleet_missions:
        debug("Exploration targets:")
        for explorationAIFleetMission in exploration_fleet_missions:
            debug(" - %s" % explorationAIFleetMission)
    else:
        debug("Exploration targets: None")

    colonisation_fleet_missions = aistate.get_fleet_missions_with_any_mission_types([MissionType.COLONISATION])
    if colonisation_fleet_missions:
        debug("Colonization targets: ")
    else:
        debug("Colonization targets: None")
    for colonisation_fleet_mission in colonisation_fleet_missions:
        debug("    %s" % colonisation_fleet_mission)

    outpost_fleet_missions = aistate.get_fleet_missions_with_any_mission_types([MissionType.OUTPOST])
    if outpost_fleet_missions:
        debug("Outpost targets: ")
    else:
        debug("Outpost targets: None")
    for outpost_fleet_mission in outpost_fleet_missions:
        debug("    %s" % outpost_fleet_mission)

    outpost_base_fleet_missions = aistate.get_fleet_missions_with_any_mission_types([MissionType.ORBITAL_OUTPOST])
    if outpost_base_fleet_missions:
        debug("Outpost Base targets (must have been interrupted by combat): ")
    else:
        debug("Outpost Base targets: None (as expected, due to expected timing of order submission and execution)")
    for outpost_fleet_mission in outpost_base_fleet_missions:
        debug("    %s" % outpost_fleet_mission)

    invasion_fleet_missions = aistate.get_fleet_missions_with_any_mission_types([MissionType.INVASION])
    if invasion_fleet_missions:
        debug("Invasion targets: ")
    else:
        debug("Invasion targets: None")
    for invasion_fleet_mission in invasion_fleet_missions:
        debug("    %s" % invasion_fleet_mission)

    troop_base_fleet_missions = aistate.get_fleet_missions_with_any_mission_types([MissionType.ORBITAL_INVASION])
    if troop_base_fleet_missions:
        debug("Invasion Base targets (must have been interrupted by combat): ")
    else:
        debug("Invasion Base targets: None (as expected, due to expected timing of order submission and execution)")
    for invasion_fleet_mission in troop_base_fleet_missions:
        debug("    %s" % invasion_fleet_mission)

    military_fleet_missions = aistate.get_fleet_missions_with_any_mission_types([MissionType.MILITARY])
    if military_fleet_missions:
        debug("General Military targets: ")
    else:
        debug("General Military targets: None")
    for military_fleet_mission in military_fleet_missions:
        debug("    %s" % military_fleet_mission)

    secure_fleet_missions = aistate.get_fleet_missions_with_any_mission_types([MissionType.SECURE])
    if secure_fleet_missions:
        debug("Secure targets: ")
    else:
        debug("Secure targets: None")
    for secure_fleet_mission in secure_fleet_missions:
        debug("    %s" % secure_fleet_mission)

    orb_defense_fleet_missions = aistate.get_fleet_missions_with_any_mission_types([MissionType.ORBITAL_DEFENSE])
    if orb_defense_fleet_missions:
        debug("Orbital Defense targets: ")
    else:
        debug("Orbital Defense targets: None")
    for orb_defence_fleet_mission in orb_defense_fleet_missions:
        debug("    %s" % orb_defence_fleet_mission)

    fleet_missions = list(aistate.get_all_fleet_missions())
    destroyed_objects = fo.getUniverse().destroyedObjectIDs(fo.empireID())

    # merge fleets where appropriate before generating fleet orders.
    # This allows us to consider the full fleet strength when determining
    # e.g. whether to engage or avoid an enemy.
    for mission in fleet_missions:
        fleet_id = mission.fleet.id
        fleet = mission.fleet.get_object()
        if not fleet or not fleet.shipIDs or fleet_id in destroyed_objects:
            continue
        mission.check_mergers()
    # get new set of fleet missions without fleets that are empty after merge
    fleet_missions = aistate.get_all_fleet_missions()
    for mission in fleet_missions:
        mission.generate_fleet_orders()


def issue_fleet_orders_for_fleet_missions():
    """Issues fleet orders."""
    debug("")
    universe = fo.getUniverse()
    aistate = get_aistate()
    fleet_missions = list(aistate.get_all_fleet_missions())
    thisround = 0
    while thisround < 3:
        thisround += 1
        debug("Issuing fleet orders round %d:" % thisround)
        for mission in fleet_missions:
            fleet_id = mission.fleet.id
            fleet = mission.fleet.get_object()
            # check that fleet was merged into another previously during this turn
            if not fleet or not fleet.shipIDs or fleet_id in universe.destroyedObjectIDs(fo.empireID()):
                continue
            mission.issue_fleet_orders()
        fleet_missions = aistate.misc.get("ReassignedFleetMissions", [])
        aistate.misc["ReassignedFleetMissions"] = []
    debug("")


def _print_systems_and_supply(system_ids):
    universe = fo.getUniverse()
    empire = fo.getEmpire()
    fleet_supplyable_system_ids = empire.fleetSupplyableSystemIDs
    for system_id in system_ids:
        system = universe.getSystem(system_id)
        debug(
            "  {}{}".format(
                system if system else "  S_%s<>" % system_id,
                "supplied" if system_id in fleet_supplyable_system_ids else "",
            )
        )


def get_fighter_capacity_of_fleet(fleet_id: int) -> tuple[int, int]:
    """
    Return current and max fighter capacity.
    """
    universe = fo.getUniverse()
    fleet = universe.getFleet(fleet_id)
    cur_capacity = 0
    max_capacity = 0
    ships = (universe.getShip(ship_id) for ship_id in (fleet.shipIDs if fleet else []))
    for ship in ships:
        design = ship and ship.design
        design_parts = design.parts if design and design.hasFighters else []
        for partname in design_parts:
            part = get_ship_part(partname)
            if part and part.partClass == fo.shipPartClass.fighterHangar:
                cur_capacity += ship.currentPartMeterValue(fo.meterType.capacity, partname)
                max_capacity += ship.currentPartMeterValue(fo.meterType.maxCapacity, partname)
    return cur_capacity, max_capacity


def get_fleet_upkeep():
    # TODO: Use new upkeep calculation
    return 1 + AIDependencies.SHIP_UPKEEP * get_aistate().shipCount


def calculate_estimated_time_of_arrival(fleet_id, target_system_id):
    universe = fo.getUniverse()
    fleet = universe.getFleet(fleet_id)
    if not fleet or not fleet.speed:
        return 99999
    distance = get_shortest_distance(fleet_id, target_system_id)
    return math.ceil(float(distance) / fleet.speed)


def get_fleet_system(fleet: Union[TargetFleet, int]) -> int:
    """Return the current fleet location or the target system if currently on starlane."""
    if isinstance(fleet, int):
        fleet = fo.getUniverse().getFleet(fleet)
    return fleet.systemID if fleet.systemID != INVALID_ID else fleet.nextSystemID


def get_current_and_max_structure(fleet: int) -> tuple[float, float]:
    """Return a 2-tuple of the sums of structure and maxStructure meters of all ships in the fleet."""

    universe = fo.getUniverse()
    destroyed_ids = universe.destroyedObjectIDs(fo.empireID())
    fleet = universe.getFleet(fleet)
    if not fleet:
        return (0.0, 0.0)
    ships_cur_health = 0
    ships_max_health = 0
    for ship_id in fleet.shipIDs:
        # Check if we have see this ship get destroyed in a different fleet since the last time we saw the subject fleet
        # this may be redundant with the new fleet assignment check made below, but for its limited scope it may be more
        # reliable, in that it does not rely on any particular handling of post-destruction stats
        if ship_id in destroyed_ids:
            continue
        this_ship = universe.getShip(ship_id)
        # check that the ship has not been seen in a new fleet since this current fleet was last observed
        if not (this_ship and this_ship.fleetID == fleet.id):
            continue
        ships_cur_health += this_ship.initialMeterValue(fo.meterType.structure)
        ships_max_health += this_ship.initialMeterValue(fo.meterType.maxStructure)

    return ships_cur_health, ships_max_health



# a list of trusted classes - other classes will not be loaded
from typing import Optional

import AIFleetMission
import AIstate
import character.character_module
import ColonisationAI
import CombatRatingsAI
import fleet_orders


class TrustedClasses:
    def __init__(self, extra_classes: Optional[dict] = None):
        self._trusted_classes = {
            f"{cls.__module__}.{cls.__name__}": cls
            for cls in [
                AIFleetMission.AIFleetMission,
                fleet_orders.AIFleetOrder,
                fleet_orders.OrderMilitary,
                fleet_orders.OrderDefend,
                fleet_orders.OrderColonize,
                fleet_orders.OrderOutpost,
                fleet_orders.OrderPause,
                fleet_orders.OrderInvade,
                fleet_orders.OrderMove,
                fleet_orders.OrderRepair,
                fleet_orders.OrderResupply,
                character.character_module.Trait,
                character.character_module.Aggression,
                character.character_module.EmpireIDTrait,
                character.character_module.Character,
                AIstate.AIstate,
                ColonisationAI.OrbitalColonizationManager,
                ColonisationAI.OrbitalColonizationPlan,
                CombatRatingsAI.ShipCombatStats,
            ]
        }
        if extra_classes:
            self._trusted_classes.update(extra_classes)

    def __contains__(self, name: str):
        return name in self._trusted_classes

    def get(self, name: str):
        return self._trusted_classes.get(name)


# prefixes to encode types not supported by json
# or not fully supported as dictionary key
ENUM_PREFIX = "__ENUM__"
INT_PREFIX = "__INT__"
FLOAT_PREFIX = "__FLOAT__"
TRUE = "__TRUE__"
FALSE = "__FALSE__"
NONE = "__NONE__"
SET_PREFIX = "__SET__"
TUPLE_PREFIX = "__TUPLE__"


# placeholder char to represent quotes in nested containers
# which would break json decoding if present.
PLACEHOLDER = "$"


class CanNotSaveGameException(Exception):
    """Exception raised when constructing the savegame string failed."""

    pass


class InvalidSaveGameException(Exception):
    """Exception raised if the savegame could not be loaded."""

    pass



from focs._effects import (
    AsteroidsType,
    Barren,
    Contains,
    Destroy,
    EffectsGroup,
    GasGiantType,
    GenerateSitRepMessage,
    IsBuilding,
    IsSource,
    Object,
    OwnedBy,
    OwnerHasTech,
    Planet,
    Population,
    SetPlanetType,
    SetPopulation,
    SetSpecies,
    Source,
    Target,
)
from macros.base_prod import BUILDING_COST_MULTIPLIER
from macros.enqueue import ENQUEUE_ARTIFICIAL_PLANET_EXCLUSION
from macros.priorities import POPULATION_OVERRIDE_PRIORITY
from macros.upkeep import COLONY_UPKEEP_MULTIPLICATOR

try:
    from focs._buildings import *
except ModuleNotFoundError:
    pass

BuildingType(  # type: ignore[reportUnboundVariable]
    name="BLD_ART_FACTORY_PLANET",
    description="BLD_ART_FACTORY_PLANET_DESC",
    buildcost=200 * Target.HabitableSize + (70 * COLONY_UPKEEP_MULTIPLICATOR * BUILDING_COST_MULTIPLIER),
    buildtime=12,
    location=(
        Planet()
        & ~Contains(IsBuilding(name=["BLD_ART_PLANET"]))
        & ~Contains(IsBuilding(name=["BLD_ART_FACTORY_PLANET"]))
        & ~Contains(IsBuilding(name=["BLD_ART_PARADISE_PLANET"]))
        & OwnedBy(empire=Source.Owner)
        & Planet(type=[AsteroidsType, GasGiantType])
        & OwnerHasTech(name="PRO_EXOBOTS")
        & Population(high=0)
    ),
    enqueuelocation=ENQUEUE_ARTIFICIAL_PLANET_EXCLUSION,
    effectsgroups=[
        EffectsGroup(
            scope=Object(id=Source.PlanetID) & Planet(),
            priority=POPULATION_OVERRIDE_PRIORITY,
            effects=[
                SetPlanetType(type=Barren),
                SetSpecies(name="SP_EXOBOT"),
                SetPopulation(value=1),
                GenerateSitRepMessage(
                    message="EFFECT_ART_PLANET",
                    label="EFFECT_ART_PLANET_LABEL",
                    icon="icons/species/robotic-01.png",
                    parameters={
                        "planet": Target.ID,
                    },
                    empire=Source.Owner,
                ),
            ],
        ),
        EffectsGroup(scope=IsSource, effects=Destroy),
    ],
    icon="icons/species/robotic-01.png",
)



"""
The TechsListAI module provides functions that describes dependencies between
various technologies to help the AI decide which technologies should be
researched next.
"""
import freeOrionAIInterface as fo
from collections.abc import Iterator
from itertools import zip_longest
from logging import debug, warning
from typing import Union

import AIDependencies as Dep
from freeorion_tools import chat_human


def unusable_techs():
    """
    Returns a list of technologies that shouldn't be researched by the AI.
    """
    return []


def defense_techs_1():
    """
    Returns a list of technologies that implement the first planetary defensive
    measures.
    """
    return ["DEF_DEFENSE_NET_1", "DEF_GARRISON_1"]


class TechGroup:
    """Base class for Tech groups.

    A TechGroup consists of some techs which need to be researched before progressing to the next TechGroup.
    The techs are split into different categories, techs within each category are researched according to order.
    Different TechGroup variants have different orders in which the tech groups are accessed.
    """

    def __init__(self):
        self.economy = []  # list of economy techs to research
        self.weapon = []  # ship weapon techs
        self.armor = []  # ship armour techs
        self.misc = []  # techs not in any other category
        self.defense = []  # planetary defense techs
        self.hull = []  # new hull types

        self._tech_queue = []  # defines the research order - forced to contain all techs.
        self._errors = []  # exceptions that occured when trying to pop from already empty lists

    @staticmethod
    def _iterate_over_remaining_techs(list_of_tech_lists: list[list[str]]) -> Iterator[str]:
        """
        Iterate over items in a stable way.

        Take 0 item from all list, after take 1 items, etc. Missed values ignored.

        >>> list(TechGroup._iterate_over_remaining_techs([["a", "b"], ["x", "y", "z"]]))
        ["a", "x", "b", "y", "z"]
        """
        return (tech for level in zip_longest(*list_of_tech_lists) for tech in level if tech)

    def _add_remaining(self):
        """Add all remaining techs in the tech group to self._tech_queue if not already contained."""
        all_lists = [self.economy, self.weapon, self.armor, self.misc, self.defense, self.hull]
        for tech in self._iterate_over_remaining_techs(all_lists):
            if tech not in self._tech_queue:
                self._tech_queue.append(tech)

    def get_techs(self) -> list:
        """Get the ordered list of techs defining research order.

        :return: Research order
        """
        self._add_remaining()
        return list(self._tech_queue)

    def enqueue(self, *tech_lists: Union[list[str], str]):
        """
        Pop first entry in the list or take entry if it is string and add it to research orders.

        Note that the passed list is modified within this function!
        If the list is already empty, the exception is caught and stored in self.__errors.
        Errors may be queried via get_errors()
        """
        for step, this_list in enumerate(tech_lists, start=1):
            if isinstance(this_list, str):
                tech_name = this_list
            else:
                try:
                    tech_name = this_list.pop(0)
                except IndexError:
                    # Do not display error message as those should be shown only once per game session
                    # by the initial test_tech_integrity() call.
                    msg = "[step %s]: Try to enqueue tech from empty list" % step
                    self._errors.append(msg)
                    continue
            if tech_name in self._tech_queue:
                msg = f"[step {step}]: Tech is already in queue: {tech_name}"
                self._errors.append(msg)
            else:
                self._tech_queue.append(tech_name)

    def get_errors(self) -> list[Exception]:
        """
        Return a list of occured exceptions.
        """
        retval = list(self._errors)
        self._errors = []
        return retval


class TechGroup1(TechGroup):
    def __init__(self):
        super().__init__()
        self.economy.extend(
            [
                "LRN_PHYS_BRAIN",
                "GRO_PLANET_ECOL",
                "GRO_SUBTER_HAB",
                "LRN_ALGO_ELEGANCE",
                Dep.LRN_ARTIF_MINDS_1,
                "PRO_ROBOTIC_PROD",
            ]
        )
        self.weapon.extend(
            [
                "SHP_WEAPON_ARC_DISRUPTOR_1",
                "SHP_WEAPON_1_2",
                "SHP_WEAPON_1_3",
                "SHP_FIGHTERS_1",
                "SHP_WEAPON_1_4",
            ]
        )
        self.defense.extend(
            [
                "DEF_GARRISON_1",
                "DEF_DEFENSE_NET_1",
            ]
        )
        self.hull.extend(
            [
                "SHP_MIL_ROBO_CONT",
                "SHP_ORG_HULL",
            ]
        )
        # always start with the same first 8 techs; leaves 1 econ, 4 weap, 2 hull
        self.enqueue(
            self.economy,
            self.economy,
            self.economy,
            self.economy,
            self.economy,
            self.weapon,
            self.defense,
            self.defense,
        )


class TechGroup1a(TechGroup1):
    def __init__(self):
        super().__init__()
        self.enqueue(
            self.weapon,
            self.weapon,
            self.weapon,
            self.weapon,
            self.economy,
            self.hull,
            "CON_ARCH_PSYCH",
        )


class TechGroup1b(TechGroup1):
    def __init__(self):
        super().__init__()
        self.enqueue(
            self.weapon,
            self.hull,
            self.economy,
            self.weapon,
            self.weapon,
            self.weapon,
            "CON_ARCH_PSYCH",
        )


class TechGroup1SparseA(TechGroup1):
    def __init__(self):
        super().__init__()
        self.enqueue(
            self.economy,
            self.hull,
            self.weapon,
            self.weapon,
            self.weapon,
            "SHP_SPACE_FLUX_DRIVE",
            self.weapon,
            "CON_ARCH_PSYCH",
        )


class TechGroup1SparseB(TechGroup1):
    def __init__(self):
        super().__init__()
        self.enqueue(
            self.economy,
            self.weapon,
            self.hull,
            "SHP_ZORTRIUM_PLATE",
            self.weapon,
            "PRO_NANOTECH_PROD",
            Dep.PRO_AUTO_1,
            "CON_ARCH_PSYCH",
            "CON_ASYMP_MATS",
            "PRO_EXOBOTS",
            "CON_ORBITAL_CON",  # not a economy tech in the strictest sense but bonus supply often equals more planets
            "GRO_GENETIC_MED",
            "GRO_SYMBIOTIC_BIO",
            "PRO_MICROGRAV_MAN",
            "GRO_XENO_GENETICS",
            "LRN_FORCE_FIELD",
            "SHP_ASTEROID_HULLS",
            "PRO_FUSION_GEN",
            "SHP_WEAPON_2_1",
            self.hull,
            "SHP_MULTICELL_CAST",
            "SHP_DEUTERIUM_TANK",
            "SHP_WEAPON_2_2",
            "PRO_ORBITAL_GEN",
            "SPY_DETECT_2",
            "SHP_SPACE_FLUX_DRIVE",
        )
        self.weapon = []


class TechGroup1SparseC(TechGroup1):
    def __init__(self):
        super().__init__()
        self.enqueue(
            self.economy,
            "SHP_ORG_HULL",
            "SHP_MULTICELL_CAST",
            "PRO_NANOTECH_PROD",
            Dep.PRO_AUTO_1,
            self.weapon,
            "CON_ARCH_PSYCH",
            "CON_ASYMP_MATS",
            "CON_ORBITAL_CON",  # not a economy tech in the strictest sense but bonus supply often equals more planets
            "GRO_GENETIC_MED",
            "GRO_SYMBIOTIC_BIO",
            "PRO_EXOBOTS",
            "SHP_DEUTERIUM_TANK",  # more reach for more scouting and planets
            "PRO_MICROGRAV_MAN",
            "GRO_XENO_GENETICS",
            "SHP_ASTEROID_HULLS",
            "PRO_FUSION_GEN",
            "SHP_WEAPON_2_1",
            "SHP_ZORTRIUM_PLATE",
            "SHP_MIL_ROBO_CONT",
            "LRN_FORCE_FIELD",
            "SHP_WEAPON_2_2",
            "PRO_ORBITAL_GEN",
            "SPY_DETECT_2",
            "SHP_SPACE_FLUX_DRIVE",
        )
        self.weapon = []


class TechGroup2(TechGroup):
    def __init__(self):
        super().__init__()
        self.economy.extend(
            [
                "PRO_FUSION_GEN",
                Dep.PRO_AUTO_1,
                "PRO_EXOBOTS",
                "GRO_SYMBIOTIC_BIO",
                "CON_ORBITAL_CON",  # not a economy tech in the strictest sense but bonus supply often equals more planets
                # "PRO_MICROGRAV_MAN",  # handled by fast-forwarding when we have asteroids
                # "PRO_ORBITAL_GEN",    # handled by fast-forwarding when we have a GG
            ]
        )
        self.armor.extend(
            [
                "SHP_ZORTRIUM_PLATE",
            ]
        )
        self.hull.extend(
            [
                "SHP_MULTICELL_CAST",
                "SHP_SPACE_FLUX_DRIVE",
                # "SHP_ASTEROID_HULLS",  # should be handled by fast-forwarding when having ASteroids
                # "SHP_DOMESTIC_MONSTER",  # should be handled by fast-forwarding when having nest
            ]
        )
        self.weapon.extend(
            [
                "SHP_WEAPON_2_1",
                "SHP_WEAPON_2_2",  # enables the building of laser based ships in ShipDesigner algorithm
                "SHP_FIGHTERS_2",
                "SHP_WEAPON_2_3",
                "SHP_WEAPON_2_4",
            ]
        )


class TechGroup2A(TechGroup2):
    def __init__(self):
        super().__init__()
        self.defense.extend(
            [
                "SPY_DETECT_2",
                "DEF_GARRISON_2",
                "LRN_FORCE_FIELD",
            ]
        )
        self.enqueue(
            self.armor,
            self.economy,
            self.economy,
            self.economy,
            self.defense,
            self.defense,
            self.hull,
            self.weapon,
            self.weapon,
            self.defense,
            self.economy,
            self.economy,
            self.weapon,
            self.weapon,
            self.weapon,
            self.hull,
        )


class TechGroup2B(TechGroup2):
    def __init__(self):
        super().__init__()
        self.defense.extend(
            [
                "LRN_FORCE_FIELD",
                "SPY_DETECT_2",
                "DEF_GARRISON_2",
            ]
        )
        self.enqueue(
            self.armor,
            self.economy,
            self.economy,
            self.defense,
            self.hull,
            self.weapon,
            self.weapon,
            self.economy,
            self.defense,
            self.defense,
            self.weapon,
            self.weapon,
            self.weapon,
            self.economy,
            self.economy,
            self.hull,
        )


class TechGroup2SparseA(TechGroup2):
    def __init__(self):
        super().__init__()
        self.enqueue(
            self.armor,
            self.hull,
            self.economy,
            self.economy,
            self.economy,
            self.economy,
            self.economy,
            Dep.LRN_ARTIF_MINDS_1,
            self.weapon,
            self.weapon,
            self.weapon,
            self.weapon,
            self.weapon,
            self.hull,
        )


class TechGroup2SparseB(TechGroup2):
    def __init__(self):
        super().__init__()
        self.enqueue(
            self.armor,
            self.hull,
            self.economy,
            self.economy,
            self.economy,
            self.economy,
            self.economy,
            Dep.LRN_ARTIF_MINDS_1,
            self.weapon,
            self.weapon,
            self.weapon,
            self.weapon,
            self.weapon,
            self.hull,
        )


class TechGroup3(TechGroup):
    def __init__(self):
        super().__init__()
        self.hull.extend(
            [
                "SHP_ASTEROID_REFORM",
                "SHP_HEAVY_AST_HULL",
                "SHP_CONTGRAV_MAINT",
            ]
        )
        self.economy.extend(
            [
                "PRO_INDUSTRY_CENTER_I",
                "GRO_GENETIC_ENG",
                "GRO_GENETIC_MED",
                "GRO_XENO_GENETICS",
                "LRN_QUANT_NET",
                "PRO_SOL_ORB_GEN",
                Dep.PRO_AUTO_2,
                "PRO_INDUSTRY_CENTER_II",
                "GRO_XENO_HYBRIDS",
                "CON_ORBITAL_HAB",
                "CON_NDIM_STRC",
                "GRO_LIFECYCLE_MAN",
            ]
        )
        self.defense.extend(
            [
                "DEF_DEFENSE_NET_2",
                "DEF_DEFENSE_NET_REGEN_1",
                "DEF_PLAN_BARRIER_SHLD_1",
                "DEF_GARRISON_3",
                "SPY_DETECT_3",
                "DEF_DEFENSE_NET_REGEN_2",
                "DEF_PLAN_BARRIER_SHLD_2",
                "DEF_DEFENSE_NET_3",
                "DEF_SYST_DEF_MINE_1",
                "DEF_PLAN_BARRIER_SHLD_3",
            ]
        )
        self.misc.extend(
            [
                "SHP_DEUTERIUM_TANK",
                "SHP_BASIC_DAM_CONT",
                "SHP_INTSTEL_LOG",
                "SHP_FLEET_REPAIR",
                "SHP_IMPROVED_ENGINE_COUPLINGS",
                "SHP_REINFORCED_HULL",
                "SHP_DEFLECTOR_SHIELD",
                "CON_CONTGRAV_ARCH",
                "CON_FRC_ENRG_STRC",
                "SHP_N_DIMENSIONAL_ENGINE_MATRIX",
            ]
        )
        self.armor.extend(
            [
                "SHP_DIAMOND_PLATE",
            ]
        )
        self.weapon.extend(
            [
                "SHP_WEAPON_3_1",
                "SHP_WEAPON_3_2",
                "SHP_FIGHTERS_3",
                "SHP_WEAPON_3_3",
                "SHP_WEAPON_3_4",
            ]
        )


class TechGroup3A(TechGroup3):
    def __init__(self):
        super().__init__()
        self.enqueue(
            self.hull,
            self.economy,
            self.misc,
            self.defense,
            self.defense,
            self.economy,
            self.economy,
            self.economy,
            self.defense,
            self.misc,
            self.defense,
            self.economy,
            self.misc,
            self.misc,
            self.economy,
            self.misc,
            self.hull,
            self.economy,
            self.hull,
            self.misc,
            self.economy,
            self.economy,
            self.defense,
            self.misc,
            self.armor,
            self.defense,
            self.defense,
            self.defense,
            self.misc,
            self.economy,
            self.misc,
            self.defense,
            self.defense,
            self.economy,
            self.misc,
            self.economy,
            self.weapon,
            self.weapon,
            self.weapon,
            self.weapon,
            self.weapon,
        )


class TechGroup3B(TechGroup3):
    def __init__(self):
        super().__init__()
        self.enqueue(
            self.hull,
            self.economy,
            self.misc,
            self.defense,
            self.defense,
            self.economy,
            self.economy,
            self.economy,
            self.weapon,
            self.weapon,
            self.defense,
            self.misc,
            self.defense,
            self.weapon,
            self.economy,
            self.misc,
            self.misc,
            self.economy,
            self.misc,
            self.weapon,
            self.weapon,
            self.hull,
            self.economy,
            self.hull,
            self.misc,
            self.economy,
            self.economy,
            self.defense,
            self.misc,
            self.armor,
            self.defense,
            self.defense,
            self.defense,
            self.misc,
            self.economy,
            self.misc,
            self.defense,
            self.defense,
            self.economy,
            self.misc,
            self.economy,
        )


class TechGroup3Sparse(TechGroup3):
    def __init__(self):
        super().__init__()
        self.enqueue(
            self.hull,
            self.misc,
            self.defense,
            self.defense,
            self.economy,
            self.economy,
            self.economy,
            self.economy,
            self.misc,
            self.economy,
            self.weapon,
            self.weapon,
            self.defense,
            self.economy,
            self.defense,
            self.economy,
            self.weapon,
            self.misc,
            self.misc,
            self.economy,
            self.misc,
            self.weapon,
            self.weapon,
            self.hull,
            self.hull,
            self.misc,
            self.economy,
            self.economy,
            self.defense,
            self.misc,
            self.armor,
            self.defense,
            self.defense,
            self.defense,
            self.misc,
            self.economy,
            self.misc,
            self.defense,
            self.defense,
            self.economy,
            self.misc,
        )


class TechGroup4(TechGroup):
    def __init__(self):
        super().__init__()
        self.hull.extend(
            [
                "SHP_FRC_ENRG_COMP",
                "SHP_MASSPROP_SPEC",
                "SHP_SCAT_AST_HULL",
            ]
        )
        self.enqueue(
            self.hull,
            self.hull,
            "SHP_ANTIMATTER_TANK",
        )


class TechGroup5(TechGroup):
    def __init__(self):
        super().__init__()
        self._tech_queue.extend(
            [
                "DEF_GARRISON_4",
                "DEF_PLAN_BARRIER_SHLD_4",
                "LRN_XENOARCH",
                "LRN_GRAVITONICS",
                "LRN_STELLAR_TOMOGRAPHY",
                "LRN_ENCLAVE_VOID",
                "SHP_MONOMOLEC_LATTICE",
                "SHP_ADV_DAM_CONT",
                "LRN_TIME_MECH",
                "SPY_DETECT_4",
                "SHP_CONT_SYMB",
                "SHP_MONOCELL_EXP",
                "GRO_TERRAFORM",
                "GRO_ENERGY_META",
                "LRN_DISTRIB_THOUGHT",
                "SHP_WEAPON_4_1",
                "SHP_WEAPON_4_2",
                "SHP_FIGHTERS_4",
                "SHP_WEAPON_4_3",
                "PRO_INDUSTRY_CENTER_III",
                "PRO_SINGULAR_GEN",
                "SHP_QUANT_ENRG_MAG",
                "SHP_PLASMA_SHIELD",
                "SHP_ENRG_BOUND_MAN",
                "SHP_WEAPON_4_4",
                "PRO_NEUTRONIUM_EXTRACTION",
                "SHP_SOLAR_CONT",
                "CON_CONC_CAMP",
                "LRN_ART_BLACK_HOLE",
                "DEF_SYST_DEF_MINE_2",
                "DEF_SYST_DEF_MINE_3",
                "LRN_PSY_DOM",
                "SHP_BLACKSHIELD",
                "DEF_PLAN_BARRIER_SHLD_5",
                "SPY_DETECT_5",
                "GRO_GAIA_TRANS",
                "CON_ART_PLANET",
            ]
        )


def _is_first_ai():
    """
    Return true if AI is first in list of AI.

    This allows to report errors in chat only by a single AI.
    """
    ais = [x for x in fo.allPlayerIDs() if not fo.playerIsHost(x)]
    return ais[0] == fo.playerID()


def _report_tech_integrity_error(message):
    if _is_first_ai():
        chat_human(message, send_to_logs=False)
    warning(message)


def test_tech_integrity():
    """Check the TechGroups for integrity.

    Try to get all tech lists by querying all the TechGroups.
    Any error is displayed in chat window, before empire greetings.
    Also checks if all techs exist and displays error if invalid tech name encountered.
    """
    tech_groups = [
        TechGroup,
        TechGroup1,
        TechGroup1a,
        TechGroup1b,
        TechGroup1SparseA,
        TechGroup1SparseB,
        TechGroup1SparseC,
        TechGroup2,
        TechGroup2A,
        TechGroup2B,
        TechGroup2SparseA,
        TechGroup2SparseB,
        TechGroup3,
        TechGroup3A,
        TechGroup3B,
        TechGroup3Sparse,
        TechGroup4,
        TechGroup5,
    ]
    debug("Checking TechGroup integrity...")
    for group in tech_groups:
        debug("Checking %s: " % group.__name__)
        error_occured = False
        this_group = group()
        techs = this_group.get_techs()
        for tech in techs:
            if not fo.getTech(tech):
                _report_tech_integrity_error(f"In {group.__name__}: Tech {tech} seems not to exist!")
                error_occured = True
        for err in this_group.get_errors():
            _report_tech_integrity_error(err)
            error_occured = True
        if not error_occured:
            debug("Seems to be OK!")


def sparse_galaxy_techs(index):
    # return primary_meta_techs()
    result = []
    debug("Choosing Sparse Research Techlist Index %d" % index)
    if index == 0:
        result = TechGroup1a().get_techs()  # early org_hull
        result += TechGroup2A().get_techs()  # prioritizes growth & defense over weapons
        result += TechGroup3A().get_techs()
        result += TechGroup4().get_techs()
        result += TechGroup5().get_techs()  #
    elif index == 1:
        result = TechGroup1b().get_techs()  # early lrn_nascent_ai
        result += TechGroup2B().get_techs()  # prioritizes growth & defense over weapons
        result += TechGroup3B().get_techs()
        result += TechGroup4().get_techs()
        result += TechGroup5().get_techs()  #
    elif index == 2:
        result = TechGroup1SparseA().get_techs()  # early lrn_nascent_ai
        result += TechGroup2SparseA().get_techs()  # prioritizes growth & defense over weapons
        result += TechGroup3Sparse().get_techs()
        result += TechGroup4().get_techs()
        result += TechGroup5().get_techs()  #
    elif index == 3:
        result = TechGroup1SparseB().get_techs()  # early org_hull
        result += TechGroup2SparseB().get_techs()
        result += TechGroup3A().get_techs()
        result += TechGroup4().get_techs()
        result += TechGroup5().get_techs()  #
    elif index == 4:
        result = TechGroup1SparseC().get_techs()  # early pro_adaptive_auto
        result += TechGroup2SparseB().get_techs()
        result += TechGroup3B().get_techs()  # faster plasma weaps
        result += TechGroup4().get_techs()
        result += TechGroup5().get_techs()  #
    return result


def primary_meta_techs(index=0):
    """
    Primary techs for all categories.
    """
    # index = 1 - index
    result = []

    debug("Choosing Research Techlist Index %d" % index)
    if index == 0:
        result = TechGroup1a().get_techs()  # early org_hull
        result += TechGroup2A().get_techs()  # prioritizes growth & defense over weapons
        result += TechGroup3A().get_techs()
        result += TechGroup4().get_techs()
        result += TechGroup5().get_techs()  #
    elif index == 1:
        result = TechGroup1b().get_techs()  # early org_hull
        result += TechGroup2B().get_techs()  # prioritizes growth & defense over weapons
        result += TechGroup3B().get_techs()
        result += TechGroup4().get_techs()
        result += TechGroup5().get_techs()  #
    elif index == 2:
        result = TechGroup1a().get_techs()  # early org_hull
        result += TechGroup2B().get_techs()  # prioritizes growth & defense over weapons
        result += TechGroup3A().get_techs()
        result += TechGroup4().get_techs()
        result += TechGroup5().get_techs()  #
    elif index == 3:
        result = TechGroup1b().get_techs()  # early org_hull
        result += TechGroup2A().get_techs()  # prioritizes growth & defense over weapons
        result += TechGroup3B().get_techs()
        result += TechGroup4().get_techs()
        result += TechGroup5().get_techs()  #
    elif index == 4:
        result = TechGroup1a().get_techs()  # early org_hull
        result += TechGroup2A().get_techs()  # prioritizes growth & defense over weapons
        result += TechGroup3B().get_techs()
        result += TechGroup4().get_techs()
        result += TechGroup5().get_techs()  #
    return result


# the following is just for reference
# "CON_ARCH_MONOFILS",
# "CON_ARCH_PSYCH",
# "CON_ART_HEAVENLY",
# "CON_ART_PLANET",
# "CON_ASYMP_MATS",
# "CON_CONC_CAMP",
# "CON_CONTGRAV_ARCH",
# "CON_FRC_ENRG_CAMO",
# "CON_FRC_ENRG_STRC",
# "CON_NDIM_STRC",
# "CON_ORBITAL_CON",
# "CON_ORBITAL_HAB",
# "CON_ORGANIC_STRC",
# "CON_PLANET_DRIVE",
# "CON_STARGATE",
# "CON_TRANS_ARCH",
# "DEF_DEFENSE_NET_1",
# "DEF_DEFENSE_NET_2",
# "DEF_DEFENSE_NET_3",
# "DEF_DEFENSE_NET_REGEN_1",
# "DEF_DEFENSE_NET_REGEN_2",
# "DEF_GARRISON_1",
# "DEF_GARRISON_2",
# "DEF_GARRISON_3",
# "DEF_GARRISON_4",
# "DEF_PLANET_CLOAK",
# "DEF_PLAN_BARRIER_SHLD_1",
# "DEF_PLAN_BARRIER_SHLD_2",
# "DEF_PLAN_BARRIER_SHLD_3",
# "DEF_PLAN_BARRIER_SHLD_4",
# "DEF_PLAN_BARRIER_SHLD_5",
# "DEF_ROOT_DEFENSE",
# "DEF_SYST_DEF_MINE_1",
# "DEF_SYST_DEF_MINE_2",
# "DEF_SYST_DEF_MINE_3",
# "GRO_ADV_ECOMAN",
# "GRO_BIOTERROR",
# "GRO_ENERGY_META",
# "GRO_GAIA_TRANS",
# "GRO_GENETIC_ENG",
# "GRO_GENETIC_MED",
# "GRO_LIFECYCLE_MAN",
# "GRO_NANOTECH_MED",
# "GRO_NANO_CYBERNET",
# "GRO_PLANET_ECOL",
# "GRO_SUBTER_HAB",
# "GRO_SYMBIOTIC_BIO",
# "GRO_TERRAFORM",
# "GRO_TRANSORG_SENT",
# "GRO_XENO_GENETICS",
# "GRO_XENO_HYBRIDS",
# "LRN_ALGO_ELEGANCE",
# "LRN_ART_BLACK_HOLE",
# "LRN_DISTRIB_THOUGHT",
# "LRN_ENCLAVE_VOID",
# "LRN_EVERYTHING",
# "LRN_FORCE_FIELD",
# "LRN_GATEWAY_VOID",
# "LRN_GRAVITONICS",
# "LRN_MIND_VOID",
# "LRN_NASCENT_AI", Dep.LRN_ARTIF_MINDS_1
# "LRN_NDIM_SUBSPACE",
# "LRN_OBSERVATORY_I",
# "LRN_PHYS_BRAIN",
# "LRN_PSIONICS",
# "LRN_PSY_DOM",
# "LRN_QUANT_NET",
# "LRN_SPATIAL_DISTORT_GEN",
# "LRN_STELLAR_TOMOGRAPHY",
# "LRN_TIME_MECH",
# "LRN_TRANSCEND",
# "LRN_TRANSLING_THT",
# "LRN_UNIF_CONC",
# "LRN_XENOARCH",
# "PRO_ADAPTIVE_AUTOMATION", Dep.PRO_AUTO_1
# "PRO_EXOBOTS",
# "PRO_FUSION_GEN",
# "PRO_INDUSTRY_CENTER_I",
# "PRO_INDUSTRY_CENTER_II",
# "PRO_INDUSTRY_CENTER_III",
# "PRO_MICROGRAV_MAN",
# "PRO_NANOTECH_PROD",
# "PRO_NDIM_ASSMB",
# "PRO_NEUTRONIUM_EXTRACTION",
# "PRO_ORBITAL_GEN",
# "PRO_ROBOTIC_PROD",
# "PRO_SENTIENT_AUTOMATION", Dep.PRO_AUTO_2
# "PRO_SINGULAR_GEN",
# "PRO_SOL_ORB_GEN",
# "PRO_ZERO_GEN",
# "SHP_ADV_DAM_CONT",
# "SHP_ANTIMATTER_TANK",
# "SHP_ASTEROID_HULLS",
# "SHP_ASTEROID_REFORM",
# "SHP_BASIC_DAM_CONT",
# "SHP_BIOADAPTIVE_SPEC",
# "SHP_BIOTERM",
# "SHP_BLACKSHIELD",
# "SHP_CAMO_AST_HULL",
# "SHP_CONTGRAV_MAINT",
# "SHP_CONT_BIOADAPT",
# "SHP_CONT_SYMB",
# "SHP_DEATH_SPORE",
# "SHP_DEFLECTOR_SHIELD",
# "SHP_DEUTERIUM_TANK",
# "SHP_DIAMOND_PLATE",
# "SHP_DOMESTIC_MONSTER",
# "SHP_ENDOCRINE_SYSTEMS",
# "SHP_ENDOSYMB_HULL",
# "SHP_ENRG_BOUND_MAN",
# "SHP_FLEET_REPAIR",
# "SHP_FRC_ENRG_COMP",
# "SHP_GAL_EXPLO",
# "SHP_HEAVY_AST_HULL",
# "SHP_IMPROVED_ENGINE_COUPLINGS",
# "SHP_INTSTEL_LOG",
# "SHP_MASSPROP_SPEC",
# "SHP_MIDCOMB_LOG",
# "SHP_MIL_ROBO_CONT",
# "SHP_MINIAST_SWARM",
# "SHP_MONOCELL_EXP",
# "SHP_MONOMOLEC_LATTICE",
# "SHP_MULTICELL_CAST",
# "SHP_MULTISPEC_SHIELD",
# "SHP_NANOROBO_MAINT",
# "SHP_NOVA_BOMB",
# "SHP_N_DIMENSIONAL_ENGINE_MATRIX",
# "SHP_ORG_HULL",
# "SHP_PLASMA_SHIELD",
# "SHP_QUANT_ENRG_MAG",
# "SHP_REINFORCED_HULL",
# "SHP_ROOT_AGGRESSION",
# "SHP_ROOT_ARMOR",
# "SHP_SCAT_AST_HULL",
# "SHP_SENT_HULL",
# "SHP_SINGULARITY_ENGINE_CORE",
# "SHP_SOLAR_CONT",
# "SHP_SPACE_FLUX_BASICS",
# "SHP_SPACE_FLUX_BUBBLE",
# "SHP_SPACE_FLUX_COMPOSITION",
# "SHP_SPACE_FLUX_DRIVE",
# "SHP_TRANSSPACE_DRIVE",
# "SHP_WEAPON_1_2",
# "SHP_WEAPON_1_3",
# "SHP_WEAPON_1_4",
# "SHP_WEAPON_2_1",
# "SHP_WEAPON_2_2",
# "SHP_WEAPON_2_3",
# "SHP_WEAPON_2_4",
# "SHP_WEAPON_3_1",
# "SHP_WEAPON_3_2",
# "SHP_WEAPON_3_3",
# "SHP_WEAPON_3_4",
# "SHP_WEAPON_4_1",
# "SHP_WEAPON_4_2",
# "SHP_WEAPON_4_3",
# "SHP_WEAPON_4_4",
# "SHP_XENTRONIUM_PLATE",
# "SHP_ZORTRIUM_PLATE",
# "SPY_DETECT_1",
# "SPY_DETECT_2",
# "SPY_DETECT_3",
# "SPY_DETECT_4",
# "SPY_DETECT_5",
# "SPY_DIST_MOD",
# "SPY_LIGHTHOUSE",
# "SPY_PLANET_STEALTH_MOD",
# "SPY_ROOT_DECEPTION",
# "SPY_STEALTH_1",
# "SPY_STEALTH_2",
# "SPY_STEALTH_3",
# "SPY_STEALTH_4",



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="LRN_GATEWAY_VOID",
    description="LRN_GATEWAY_VOID_DESC",
    short_description="BUILDING_UNLOCK_SHORT_DESC",
    category="LEARNING_CATEGORY",
    researchcost=240 * TECH_COST_MULTIPLIER,
    researchturns=5,
    tags=["PEDIA_LEARNING_CATEGORY"],
    prerequisites=["CON_STARGATE"],
    unlock=Item(type=UnlockBuilding, name="BLD_GATEWAY_VOID"),
    graphic="icons/tech/monument_to_exodus.png",
)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="SHP_NANOROBO_MAINT",
    description="SHP_NANOROBO_MAINT_DESC",
    short_description="SHIP_HULL_UNLOCK_SHORT_DESC",
    category="SHIP_HULLS_CATEGORY",
    researchcost=500 * TECH_COST_MULTIPLIER,
    researchturns=10,
    tags=["PEDIA_ROBOTIC_HULL_TECHS"],
    prerequisites=["SHP_MIL_ROBO_CONT", "PRO_NANOTECH_PROD"],
    unlock=[
        Item(type=UnlockShipHull, name="SH_NANOROBOTIC"),
        Item(type=UnlockBuilding, name="BLD_SHIPYARD_CON_NANOROBO"),
    ],
    graphic="icons/ship_hulls/nano_robotic_hull_small.png",
)
